import inspect
from datetime import datetime

from . import _LOGGER
from .const import TEXT_ON, TEXT_UNKNOWN, WISERSYSTEM
from .helpers.capabilities import (
    _WiserHubCapabilitiesInfo,
    _WiserHubFeatureCapabilitiesInfo,
)
from .helpers.cloud import _WiserCloud
from .helpers.firmware import _WiserFirmareUpgradeInfo
from .helpers.gps import _WiserGPS
from .helpers.network import _WiserNetwork
from .helpers.opentherm import _WiserOpentherm
from .helpers.signal import _WiserSignalStrength
from .helpers.special_times import sunrise_times, sunset_times
from .helpers.temp import _WiserTemperatureFunctions as tf
from .helpers.zigbee import _WiserZigbee
from .rest_controller import _WiserRestController


class _WiserSystem(object):
    """Class representing a Wiser Hub device"""

    def __init__(
        self,
        wiser_rest_controller: _WiserRestController,
        domain_data: dict,
        network_data: dict,
        device_data: dict,
        opentherm_data: dict,
    ):
        self._wiser_rest_controller = wiser_rest_controller
        self._data = domain_data
        self._system_data = self._data.get("System", {})

        # Sub classes for system setting values
        self._capability_data = _WiserHubCapabilitiesInfo(
            self._data.get("DeviceCapabilityMatrix", {})
        )
        self._feature_capability_data = _WiserHubFeatureCapabilitiesInfo(
            self._data.get("FeatureCapability", {})
        )
        self._cloud_data = _WiserCloud(
            self._system_data.get("CloudConnectionStatus"),
            self._data.get("Cloud", {}),
        )
        self._device_data = self._get_system_device(device_data)
        self._network_data = _WiserNetwork(
            network_data.get("Station", {}), self._wiser_rest_controller
        )
        self._opentherm_data = _WiserOpentherm(
            self._wiser_rest_controller,
            opentherm_data,
            self._system_data.get("OpenThermConnectionStatus", TEXT_UNKNOWN),
        )
        self._signal = _WiserSignalStrength(self._device_data)
        self._upgrade_data = _WiserFirmareUpgradeInfo(self._data.get("UpgradeInfo", {}))
        self._zigbee_data = _WiserZigbee(self._data.get("Zigbee", {}))

        # Variables to hold values for settabel values
        self._automatic_daylight_saving = self._system_data.get(
            "AutomaticDaylightSaving"
        )
        self._away_mode_affects_hotwater = self._system_data.get(
            "AwayModeAffectsHotWater"
        )
        self._away_mode_target_temperature = self._system_data.get(
            "AwayModeSetPointLimit"
        )
        self._comfort_mode_enabled = self._system_data.get("ComfortModeEnabled")
        self._degraded_mode_target_temperature = self._system_data.get(
            "DegradedModeSetpointThreshold"
        )
        self._eco_mode_enabled = self._system_data.get("EcoModeEnabled")
        self._hub_time = datetime.fromtimestamp(self._system_data.get("UnixTime"))
        self._override_type = self._system_data.get("OverrideType", "")
        self._timezone_offset = self._system_data.get("TimeZoneOffset")
        self._valve_protection_enabled = self._system_data.get("ValveProtectionEnabled")

        # Added by LGO
        # Summer comfort
        self._summer_comfort_enabled = self._system_data.get("SummerComfortEnabled")
        self._indoor_discomfort_temperature = self._system_data.get(
            "IndoorDiscomfortTemperature"
        )
        self._outdoor_discomfort_temperature = self._system_data.get(
            "OutdoorDiscomfortTemperature"
        )
        self._summer_comfort_available = self._system_data.get("SummerComfortAvailable")
        self._summer_discomfort_prevention = self._system_data.get(
            "SummerDiscomfortPrevention"
        )

        # End Added by LGO

    def _get_system_device(self, device_data: dict):
        for device in device_data:
            # Add controller to sytem class
            if device.get("ProductType") == "Controller":
                return device

    async def _send_command(self, cmd: dict, path: str = None) -> bool:
        """
        Send system control command to Wiser Hub
        param cmd: json command structure
        return: boolen - true = success, false = failed
        """
        if path:
            result = await self._wiser_rest_controller._send_command(
                f"{WISERSYSTEM}/{path}", cmd
            )
        else:
            result = await self._wiser_rest_controller._send_command(WISERSYSTEM, cmd)
        if result:
            _LOGGER.debug(
                "Wiser hub - {} command successful".format(inspect.stack()[1].function)
            )
            return True
        return False

    @property
    def active_system_version(self) -> str:
        """Get current hub firmware version"""
        return self._system_data.get("ActiveSystemVersion", TEXT_UNKNOWN)

    @property
    def automatic_daylight_saving_enabled(self) -> bool:
        """Get or set if auto daylight saving is enabled"""
        return self._automatic_daylight_saving

    async def set_automatic_daylight_saving_enabled(self, enabled: bool):
        if await self._send_command({"AutomaticDaylightSaving": str(enabled).lower()}):
            self._automatic_daylight_saving = enabled
            return True

    @property
    def away_mode_enabled(self) -> bool:
        """Get or set if away mode is enabled"""
        return True if self._override_type == "Away" else False

    async def set_away_mode_enabled(self, enabled: bool):
        if await self._send_command({"RequestOverride": {"Type": 2 if enabled else 0}}):
            self._override_type = "Away" if enabled else ""
            return True

    @property
    def away_mode_affects_hotwater(self) -> bool:
        """Get or set if setting away mode affects hot water"""
        return self._away_mode_affects_hotwater

    async def set_away_mode_affects_hotwater(self, enabled: bool = False):
        if await self._send_command({"AwayModeAffectsHotWater": str(enabled).lower()}):
            self._away_mode_affects_hotwater = enabled
            return True

    @property
    def away_mode_target_temperature(self) -> float:
        """Get or set target temperature for away mode"""
        return tf._from_wiser_temp(self._away_mode_target_temperature)

    async def set_away_mode_target_temperature(self, temp: float):
        temp = tf._to_wiser_temp(temp)
        if await self._send_command({"AwayModeSetPointLimit": temp}):
            self._away_mode_target_temperature = tf._to_wiser_temp(temp)
            return True

    @property
    def boiler_fuel_type(self) -> str:
        """Get boiler fuel type setting"""
        # TODO: Add ability to set to 1 of 3 types
        return self._system_data.get("BoilerSettings", {"FuelType": TEXT_UNKNOWN}).get(
            "FuelType"
        )

    @property
    def brand_name(self) -> str:
        """Get brand name of Wiser hub"""
        return self._system_data.get("BrandName")

    @property
    def capabilities(self) -> _WiserHubCapabilitiesInfo:
        """Get capability info"""
        return self._capability_data

    @property
    def cloud(self) -> _WiserCloud:
        """Get cloud settings"""
        return self._cloud_data

    @property
    def comfort_mode_enabled(self) -> bool:
        """Get or set if comfort mode is enabled"""
        return self._comfort_mode_enabled

    async def set_comfort_mode_enabled(self, enabled: bool):
        if await self._send_command({"ComfortModeEnabled": enabled}):
            self._comfort_mode_enabled = enabled
            return True

    @property
    def degraded_mode_target_temperature(self) -> float:
        """Get or set degraded mode target temperature"""
        return tf._from_wiser_temp(self._degraded_mode_target_temperature)

    async def set_degraded_mode_target_temperature(self, temp: float):
        temp = temp._to_wiser_temp(temp)
        if await self._send_command({"DegradedModeSetpointThreshold": temp}):
            self._degraded_mode_target_temperature = temp
            return True

    @property
    def eco_mode_enabled(self) -> bool:
        """Get or set whether eco mode is enabled"""
        return self._eco_mode_enabled

    async def set_eco_mode_enabled(self, enabled: bool):
        if await self._send_command({"EcoModeEnabled": enabled}):
            self._eco_mode_enabled = enabled
            return True

    @property
    def feature_capabilities(self) -> _WiserHubFeatureCapabilitiesInfo:
        """Get feature capability info"""
        return self._feature_capability_data

    @property
    def firmware_over_the_air_enabled(self) -> bool:
        """Whether firmware updates over the air are enabled on the hub"""
        return self._system_data.get("FotaEnabled")

    @property
    def firmware_version(self) -> str:
        """Get firmware version of device"""
        return self._device_data.get("ActiveFirmwareVersion", TEXT_UNKNOWN)

    @property
    def geo_position(self) -> _WiserGPS:
        """Get geo location information"""
        return _WiserGPS(self._system_data.get("GeoPosition", {}))

    @property
    def hardware_generation(self) -> int:
        """Get hardware generation version"""
        return self._system_data.get("HardwareGeneration", 1)

    @property
    def heating_button_override_state(self) -> bool:
        """Get if heating override button is on"""
        return (
            True
            if self._system_data.get("HeatingButtonOverrideState") == TEXT_ON
            else False
        )

    @property
    def hotwater_button_override_state(self) -> bool:
        """Get if hot water override button is on"""
        return (
            True
            if self._system_data.get("HotWaterButtonOverrideState") == TEXT_ON
            else False
        )

    @property
    def hub_time(self) -> datetime:
        """Get the current time on hub"""
        return self._hub_time

    @property
    def id(self) -> int:
        """Get id of device"""
        return self._device_data.get("id")

    @property
    def is_away_mode_enabled(self) -> bool:
        """Get if away mode is enabled"""
        return True if self._override_type == "Away" else False

    @property
    def model(self) -> str:
        """Get model of device"""
        return self._device_data.get("ModelIdentifier", TEXT_UNKNOWN)

    @property
    def name(self) -> str:
        """Get name of hub"""
        return self.network.hostname

    @property
    def network(self) -> _WiserNetwork:
        """Get network information from hub"""
        return self._network_data

    @property
    def node_id(self) -> int:
        """Get zigbee node id of device"""
        return self._device_data.get("NodeId", 0)

    @property
    def opentherm(self) -> _WiserOpentherm:
        """Get opentherm info"""
        return self._opentherm_data

    @property
    def pairing_status(self) -> str:
        """Get account pairing status"""
        return self._system_data.get("PairingStatus", TEXT_UNKNOWN)

    @property
    def parent_node_id(self) -> int:
        """Get zigbee node id of device this device is connected to"""
        return self._device_data.get("ParentNodeId", 0)

    @property
    def product_type(self) -> str:
        """Get product type of device"""
        return self._device_data.get("ProductType", TEXT_UNKNOWN)

    @property
    def signal(self) -> _WiserSignalStrength:
        """Get zwave network information"""
        return self._signal

    # Added LGO
    @property
    def summer_comfort_enabled(self) -> bool:
        """Get whether summer comfort mode is enabled"""
        return self._summer_comfort_enabled

    async def set_summer_comfort_enabled(self, enabled: bool):
        """Set whether summer comfort mode is enabled"""
        if await self._send_command({"SummerComfortEnabled": enabled}):
            self._summer_comfort_enabled = enabled
            return True

    @property
    def indoor_discomfort_temperature(self) -> float:
        """Get indoor discomfort temperature for summer comfort"""
        return tf._from_wiser_temp(self._indoor_discomfort_temperature)

    async def set_indoor_discomfort_temperature(self, temp: float):
        """Set indoor discomfort temperature for summer comfort"""
        temp = tf._to_wiser_temp(temp)
        if await self._send_command({"IndoorDiscomfortTemperature": temp}):
            self._away_mode_target_temperature = tf._to_wiser_temp(temp)
            return True

    @property
    def outdoor_discomfort_temperature(self) -> float:
        """Get outdoor discomfort temperature for summer comfort"""
        return tf._from_wiser_temp(self._outdoor_discomfort_temperature)

    async def set_outdoor_discomfort_temperature(self, temp: float):
        """Set outdoor discomfort temperature for summer comfort"""
        temp = tf._to_wiser_temp(temp)
        if await self._send_command({"OutdoorDiscomfortTemperature": temp}):
            self._away_mode_target_temperature = tf._to_wiser_temp(temp)
            return True

    @property
    def summer_comfort_available(self) -> bool:
        """Get whether summer comfort mode is available"""
        return self._summer_comfort_available

    @property
    def summer_discomfort_prevention(self) -> bool:
        """Get summer discomfort prevention"""
        return self._summer_discomfort_prevention

    async def set_summer_discomfort_prevention(self, enabled: bool):
        """Set summer discomfort prevention"""
        if await self._send_command({"SummerDiscomfortPrevention": enabled}):
            self._summer_discomfort_prevention = enabled
            return True

    @property
    def type_comm(self) -> str:
        """Get type of zigbee device"""
        return self._device_data.get("Type", TEXT_UNKNOWN)

    @property
    def uuid(self) -> str:
        """Get UUID zigbee"""
        return self._device_data.get("UUID", TEXT_UNKNOWN)

    # PCM features
    @property
    def pcm_version(self) -> str:
        """Get PCM version"""
        return self._system_data.get("PCMVersion", TEXT_UNKNOWN)

    @property
    def pcm_status(self) -> str:
        """Get PCM Status"""
        return self._system_data.get("PCMStatus", TEXT_UNKNOWN)

    @property
    def pcm_device_limit_reached(self) -> bool:
        """Get PCM device limit reached"""
        return self._system_data.get("PCMDeviceLimitReached", False)

    @property
    def can_activate_pcm(self) -> bool:
        """Get Can activate PCM"""
        return self._system_data.get("CanActivatePCM", False)

    # End Added LGO

    @property
    def sunrise_times(self) -> dict[str, str]:
        """Get sunrise times"""
        return sunrise_times(self._system_data.get("SunriseTimes", []))

    @property
    def sunset_times(self) -> dict[str, str]:
        """Get sunset times"""
        return sunset_times(self._system_data.get("SunsetTimes", []))

    @property
    def system_mode(self) -> str:
        """Get current system mode"""
        return self._system_data.get("SystemMode", TEXT_UNKNOWN)

    @property
    def timezone_offset(self) -> int:
        """Get timezone offset in minutes"""
        return self._timezone_offset

    async def set_timezone_offset(self, offset: int):
        if await self._send_command({"TimeZoneOffset": offset}):
            self._timezone_offset = offset
            return True

    @property
    def user_overrides_active(self) -> bool:
        """Get if any overrides are active"""
        return self._system_data.get("UserOverridesActive", False)

    @property
    def valve_protection_enabled(self) -> bool:
        """Get or set if valve protection is enabled"""
        return self._valve_protection_enabled

    async def set_valve_protection_enabled(self, enabled: bool):
        """
        Set the valve protection setting on the wiser hub
        param enabled: turn on or off
        """
        return await self._send_command({"ValveProtectionEnabled": enabled})

    @property
    def zigbee(self) -> _WiserZigbee:
        """Get zigbee info"""
        return self._zigbee_data

    async def allow_add_device(self, allow_time: int = 120):
        """
        Put hub in permit join mode for adding new devices
        """
        return await self._send_command(allow_time, "RequestPermitJoin")

    async def boost_all_rooms(self, inc_temp: float, duration: int) -> bool:
        """
        Boost the temperature of all rooms
        param inc_temp: increase target temperature over current temperature by 0C to 5C
        param duration: the duration to boost the room temperatures in minutes
        return: boolean
        """
        return await self._send_command(
            {
                "RequestOverride": {
                    "Type": "Boost",
                    "DurationMinutes": duration,
                    "IncreaseSetPointBy": tf._to_wiser_temp(inc_temp, "boostDelta"),
                }
            }
        )

    async def cancel_all_overrides(self):
        """
        Cancel all overrides and set room schedule to the current temperature setting for the mode
        return: boolean
        """
        return await self._send_command(
            {"RequestOverride": {"Type": "CancelUserOverrides"}}
        )



from .helpers.battery import _WiserBattery
from .helpers.device import _WiserDevice
from .helpers.temp import _WiserTemperatureFunctions as tf


class _WiserRoomStat(_WiserDevice):
    """Class representing a Wiser Room Stat device"""

    @property
    def battery(self) -> _WiserBattery:
        """Get the battery information for the room stat"""
        return _WiserBattery(self._data)

    @property
    def current_humidity(self) -> int:
        """Get the current humidity reading of the room stat"""
        return self._device_type_data.get("MeasuredHumidity", 0)

    @property
    def current_target_temperature(self) -> float:
        """Get the room stat current target temperature setting"""
        return tf._from_wiser_temp(self._device_type_data.get("SetPoint", 0))

    @property
    def current_temperature(self) -> float:
        """Get the current temperature measured by the room stat"""
        return tf._from_wiser_temp(
            self._device_type_data.get("MeasuredTemperature", 0), "current"
        )


class _WiserRoomStatCollection:
    """Class holding all wiser room stats"""

    def __init__(self):
        self._items = []

    @property
    def all(self) -> list[_WiserRoomStat]:
        return list(self._items)

    @property
    def count(self) -> int:
        return len(self.all)

    # Roomstats
    def get_by_id(self, roomstat_id: int) -> _WiserRoomStat:
        """
        Gets a RoomStat object from the RoomStats id
        param id: id of room stat
        return: _WiserRoomStat object
        """
        for roomstat in self.all:
            if roomstat.id == roomstat_id:
                return roomstat
        return None



from typing import Any

from .helpers.device import _WiserDevice


class _WiserButtonPanel(_WiserDevice):
    """Class representing a Wiser Button Panel"""

    @property
    def number_of_gangs(self) -> list[int]:
        """Number of gangs."""
        return self._device_type_data.get("NumberOfGangs")

    @property
    def events(self) -> list[dict[str, Any]]:
        """Events info."""
        return self._device_type_data.get("Events")


class _WiserButtonPanelCollection:
    """Class holding all Wiser Button Panels"""

    def __init__(self):
        self._items = []

    @property
    def all(self) -> list[_WiserButtonPanel]:
        return list(self._items)

    @property
    def count(self) -> int:
        return len(self.all)

    def get_by_id(self, id: int) -> _WiserButtonPanel:
        """
        Gets a button panel object from the sensor id
        param id: id of button panel
        return: _WiserButtonPanel object
        """
        try:
            return [buttonpanel for buttonpanel in self.all if buttonpanel.id == id][0]
        except IndexError:
            return None



import argparse
import json
import logging
import pathlib
from sys import version_info

from . import __VERSION__

from .const import WISERHUBDOMAIN, WISERHUBNETWORK, WISERHUBSCHEDULES
from .exceptions import (
    WiserHubAuthenticationError,
    WiserHubConnectionError,
    WiserHubRESTError,
)
from aioWiserHeatAPI.rest_controller import _WiserConnectionInfo, _WiserRestController


def main_parser() -> argparse.ArgumentParser:
    logging.basicConfig(level=logging.ERROR)

    parser = argparse.ArgumentParser(
        description="A simple executable to use and test the wiserHeatAPIv2 library."
    )
    subparsers = parser.add_subparsers(dest="cmd")
    subparsers.required = True

    output_parser = subparsers.add_parser(
        "output", description="Output json endpoints to files"
    )
    output_parser.add_argument(
        "-r",
        "--raw",
        dest="raw",
        action="store_const",
        help="(optional) By default, sensitive data is anonomised.  Set this flag to output raw data without anonomisation from hub",
        const=True,
        default=False,
    )
    _add_default_arguments(output_parser)
    output_parser.add_argument(
        "type", help="What to output. Possible values are domain/network/schedule/all"
    )
    output_parser.set_defaults(func=output_json)

    version_parser = subparsers.add_parser("version", description="Show api version")
    version_parser.set_defaults(func=show_version)

    return parser


def output_json(args) -> None:
    c = _WiserConnectionInfo()
    c.host = args.hostname
    c.secret = args.secret

    wiser_rest_controller = _WiserRestController(c)

    # Read data from hub
    try:
        if args.type == "domain" or args.type == "all":
            domain_data = wiser_rest_controller._get_hub_data(WISERHUBDOMAIN)
            log_response_to_file(domain_data, "domain", args.raw)
        if args.type == "network" or args.type == "all":
            network_data = wiser_rest_controller._get_hub_data(WISERHUBNETWORK)
            log_response_to_file(network_data, "network", args.raw)
        if args.type == "schedule" or args.type == "all":
            schedule_data = wiser_rest_controller._get_hub_data(WISERHUBSCHEDULES)
            log_response_to_file(schedule_data, "schedule", args.raw)
    except WiserHubAuthenticationError:
        print(
            "Unable to authenticate with your Wiser HeatHub.  Please check your secret key."
        )
    except WiserHubConnectionError:
        print(
            "Unable to connect to your Wiser HeatHub.  Check your ip/hostname and that your hub is responding on the network."
        )
    except WiserHubRESTError:
        print(
            "Your Wiser HeatHub returned an incorrectly formatted response.  Please check you are conncting to the correct device and try again."
        )
    except Exception as ex:
        print(
            f"Unknown error getting json data from your Wiser HeatHub.  Error is {ex}"
        )


def show_version(args) -> None:
    print(f"API version is {__VERSION__}")


def _add_default_arguments(parser: argparse.ArgumentParser):
    """Add the default arguments username, password, region to the parser."""
    parser.add_argument("hostname", help="HeatHub IP or DNS name")
    parser.add_argument("secret", help="Heathub secret key")


def log_response_to_file(
    json_data: str,
    logfilename: str,
    raw: bool,
    log_responses: pathlib.Path = pathlib.Path.home(),
) -> None:
    """If a log path is set, log all resonses to a file."""
    if logfilename is None:
        return

    if raw:
        data = json.dumps(json_data, indent=2, sort_keys=False)
    else:
        data = json.dumps(anonymise_data(json_data), indent=2, sort_keys=False)

    output_path = log_responses / "wiser_data/{}.json".format(logfilename)

    output_path.parent.mkdir(exist_ok=True, parents=True)
    with open(output_path, "w", encoding="UTF-8") as logfile:
        logfile.write(data)

    print(f"{logfilename.title()} data written to {output_path}")


def anonymise_data(json_data: dict) -> dict:
    """Replace parts of the logfiles containing personal information."""

    replacements = {
        "Latitude": 30.4865,
        "Longitude": 58.4892,
        "SerialNumber": "ANON_SERIAL",
        "MacAddress": "ANON_MAC",
        "HostName": "WiserHeatXXXXXX",
        "MdnsHostname": "WiserHeatXXXXXX",
        "IPv4Address": "ANON_IP",
        "IPv4HostAddress": "ANON_IP",
        "IPv4DefaultGateway": "ANON_IP",
        "IPv4PrimaryDNS": "ANON_IP",
        "IPv4SecondaryDNS": "ANON_IP",
        "SSID": "ANON_SSID",
        "DetectedAccessPoints": [],
    }

    if isinstance(json_data, dict):
        for key, value in json_data.items():
            if isinstance(value, dict):
                json_data[key] = anonymise_data(value)
            elif isinstance(value, list):
                key_data = []
                for item in value:
                    key_data.append(anonymise_data(item))
                json_data[key] = key_data
            elif key in replacements:
                json_data[key] = replacements[key]
    return json_data


def main():
    """Main function."""
    parser = main_parser()
    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



"""Version helpers."""

from dataclasses import dataclass


@dataclass
class Version:
    """Class to handle PEP440 versions."""

    version: str

    def __gt__(self, other):
        if isinstance(other, Version):
            return self._version_compare(self.version, other.version) == 1

    def __ge__(self, other):
        if isinstance(other, Version):
            return self.__eq__(other) or self.__gt__(other)

    def __lt__(self, other):
        if isinstance(other, Version):
            return self._version_compare(self.version, other.version) == -1

    def __le__(self, other):
        if isinstance(other, Version):
            return self.__eq__(other) or self.__lt__(other)

    def __eq__(self, other):
        if isinstance(other, Version):
            return self._version_compare(self.version, other.version) == 0

    def _version_compare(self, v1: str, v2: str) -> int:
        version1 = v1.replace("-", ".").split(".")
        version2 = v2.replace("-", ".").split(".")

        length = max(len(version1), len(version2))

        # Compare each component of the version strings.
        for i in range(length):
            v1_itm = int(version1[i]) if version1[i].isdigit() else version1[i]
            v2_itm = int(version2[i]) if version2[i].isdigit() else version2[i]

            if v1_itm > v2_itm:
                return 1
            if v1_itm < v2_itm:
                return -1

        return 0



import inspect

from .. import _LOGGER
from ..const import TEXT_UNKNOWN
from ..rest_controller import _WiserRestController
from .temp import _WiserTemperatureFunctions as tf


class _WiserThresholdSensor:
    """Class representing a Wiser threshold sensor for a WIser device."""

    def __init__(
        self,
        wiser_rest_controller: _WiserRestController,
        endpoint: str,
        data: dict,
    ):
        self._wiser_rest_controller = wiser_rest_controller
        self._endpoint = endpoint
        self._data = data

    async def _send_command(self, cmd: dict):
        """
        Send control command to the Threshold sensor
        param cmd: json command structure
        return: boolen - true = success, false = failed
        """
        result = await self._wiser_rest_controller._send_command(
            self._endpoint.format(self.id), cmd
        )
        if result:
            self._device_type_data = result
        if result:
            _LOGGER.debug(
                "Wiser light - {} command successful".format(
                    inspect.stack()[1].function
                )
            )
            return True
        return False

    @property
    def id(self) -> str:
        """Get the current humidity reading of the room stat"""
        return self._data.get("id", 0)

    @property
    def UUID(self) -> str:
        """Get the current humidity reading of the room stat"""
        return self._data.get("UUID", TEXT_UNKNOWN)

    @property
    def quantity(self) -> str:
        """Get the current humidity reading of the room stat"""
        return self._data.get("Quantity", TEXT_UNKNOWN)

    @property
    def current_value(self) -> float:
        """Get the current humidity reading of the room stat"""
        if self._data.get("Quantity") == "Temperature":
            return tf._from_wiser_temp(self._data.get("CurrentValue"), "current")
        else:
            return self._data.get("CurrentValue", 0)

    @property
    def high_threshold(self) -> float:
        """Get the high threshold of the sensor setting"""
        return self._data.get("HighThreshold", 0)

    @property
    def medium_threshold(self) -> float:
        """Get the high threshold of the sensor setting"""
        return self._data.get("MediumThreshold", 0)

    @property
    def low_threshold(self) -> float:
        """Get the high threshold of the sensor setting"""
        return self._data.get("LowThreshold", 0)

    @property
    def current_level(self) -> str:
        """Get the high threshold of the sensor setting"""
        return self._data.get("CurrentLevel", TEXT_UNKNOWN)

    @property
    def interacts_with_room_climate(self) -> bool:
        """Get the current temperature measured by the room stat"""
        return self._data.get("InteractsWithRoomClimate", False)

    async def set_interacts_with_room_climate(self, enabled: bool):
        if await self._send_command({"InteractsWithRoomClimate": str(enabled).lower()}):
            self._interacts_with_room_climate = enabled
            return True



"""Temperature helper functions."""

from ..const import (
    TEMP_ERROR,
    TEMP_HW_OFF,
    TEMP_HW_ON,
    TEMP_MAXIMUM,
    TEMP_MINIMUM,
    TEMP_OFF,
    WiserTempLimitsEnum,
    WiserUnitsEnum,
)


class _WiserTemperatureFunctions(object):
    # -----------------------------------------------------------
    # Support Functions
    # -----------------------------------------------------------
    @staticmethod
    def _to_wiser_temp(
        temp: float,
        temp_type: str = "heating",
        units: WiserUnitsEnum = WiserUnitsEnum.metric,
    ) -> float | int:
        """
        Converts from degrees C to wiser hub format
        param temp: The temperature to convert
        param type: Can be heating (default), hotwater or delta
        return: Integer
        """
        temp = int(
            _WiserTemperatureFunctions._validate_temperature(temp, temp_type) * 10
        )

        # Convert to metric if imperial units set
        if units == WiserUnitsEnum.imperial:
            temp = _WiserTemperatureFunctions._convert_from_fahrenheit(temp)

        return temp

    @staticmethod
    def _from_wiser_temp(
        temp: float | int,
        temp_type: str = "heating",
        units: WiserUnitsEnum = WiserUnitsEnum.metric,
    ) -> float | None:
        """
        Converts from wiser hub format to degrees C
        param temp: The wiser temperature to convert
        return: Float
        """
        if temp is not None:
            if abs(temp) >= TEMP_ERROR:
                # Fix high value from hub when lost sight of iTRV
                return None
            else:
                temp = _WiserTemperatureFunctions._validate_temperature(
                    round(temp / 10, 1), temp_type
                )

            # Convert to imperial if imperial units set
            if units == WiserUnitsEnum.imperial:
                temp = _WiserTemperatureFunctions._convert_to_fahrenheit(temp)

            return temp
        return None

    @staticmethod
    def _is_valid_temp(temp: float, hw: bool = False) -> bool:
        if hw and temp in [TEMP_HW_ON, TEMP_HW_OFF]:
            return True
        elif int(temp) == TEMP_OFF:
            return True
        elif temp >= TEMP_MINIMUM and temp <= TEMP_MAXIMUM:
            return True
        return False

    @staticmethod
    def _validate_temperature(temp: float, temp_type: str = "heating") -> float | None:
        """
        Validates temperature value is in range of Wiser Hub allowed values
        Sets to min or max temp if value exceeds limits
        param temp: temperature value to validate
        return: float
        """

        # Get limits type
        limits = WiserTempLimitsEnum[temp_type].value
        if limits.get("type") == "onoff":
            if temp in [limits.get("on"), limits.get("off")]:
                return temp
            return limits.get("off")

        if limits.get("type") == "range":
            if temp >= TEMP_ERROR:
                return limits.get("min")
            elif temp > limits.get("max"):
                return limits.get("max")
            elif temp < limits.get("min") and temp != limits.get("off", TEMP_OFF):
                return limits.get("min")
            else:
                return temp

        raise ValueError("Invalid temperature type for validation")

    @staticmethod
    def _convert_from_fahrenheit(temp: float) -> float:
        """
        Convert F temp to C
        param temp: temp in F to convert
        return: float
        """
        return round((temp - 32) * 5 / 9, 1)

    @staticmethod
    def _convert_to_fahrenheit(temp: float) -> float:
        """
        Convert C temp to F
        param temp: temp in C to convert
        return: float
        """
        return round((temp * 9 / 5) + 32, 1)



import asyncio
import inspect
from datetime import datetime

from aioWiserHeatAPI.exceptions import WiserExtraConfigError

from . import _LOGGER
from .const import (
    TEMP_HW_OFF,
    TEMP_HW_ON,
    TEXT_BOOST,
    TEXT_OFF,
    TEXT_ON,
    TEXT_UNKNOWN,
    WISER_BOOST_DURATION,
    WISERHOTWATER,
    WiserDeviceModeEnum,
    WiserHotWaterClimateModeEnum,
    WiserPresetOptionsEnum,
)
from .helpers.misc import is_value_in_list
from .helpers.temp import _WiserTemperatureFunctions as tf
from .rest_controller import _WiserRestController
from .schedule import _WiserSchedule

DEFAULT_TARGET_TEMP = 60


class _WiserHotwater(object):
    """Class representing a Wiser Hot Water controller"""

    def __init__(
        self,
        wiser_rest_controller: _WiserRestController,
        hw_data: dict,
        schedule: _WiserSchedule,
    ):
        self._wiser_rest_controller = wiser_rest_controller
        self._data = hw_data
        self._schedule = schedule

        self._extra_config = (
            self._wiser_rest_controller._extra_config.config("HotWater", str(self.id))
            if self._wiser_rest_controller._extra_config
            else None
        )

        self._default_extra_config = {
            "climate_off": False,
            "target_temperature_low": DEFAULT_TARGET_TEMP - 8,
            "target_temperature_high": DEFAULT_TARGET_TEMP,
            "manual_heat": False,
        }

        self._current_temperature: float = 0.0

        # Add device id to schedule
        if self._schedule:
            self.schedule._assignments.append({"id": self.id, "name": self.name})

    async def _send_command(self, cmd: dict):
        """
        Send control command to the hot water
        param cmd: json command structure
        return: boolen - true = success, false = failed
        """
        result = await self._wiser_rest_controller._send_command(
            WISERHOTWATER.format(self.id), cmd
        )
        if result:
            _LOGGER.debug(
                "Wiser hot water - {} command successful".format(
                    inspect.stack()[1].function
                )
            )
            return True
        return False

    async def _update_extra_config(self, key: str, value) -> bool:
        if self._wiser_rest_controller._extra_config:
            try:
                await self._wiser_rest_controller._extra_config.async_update_config(
                    "HotWater", str(self.id), {key: value}
                )
                await self._get_extra_config()
                return True
            except Exception:
                return False
        return False

    async def _get_extra_config(self) -> None:
        self._extra_config = (
            self._wiser_rest_controller._extra_config.config("HotWater", str(self.id))
            if self._wiser_rest_controller._extra_config
            else self._default_extra_config
        )

    @property
    def available_modes(self) -> list[str]:
        """Return available HVAC modes."""
        if self.is_climate_mode:
            return [mode.value for mode in WiserHotWaterClimateModeEnum]
        return [mode.value for mode in WiserDeviceModeEnum]

    @property
    def available_presets(self) -> list:
        """Get available preset modes"""
        # Remove advance schedule if no schedule exists or in climate mode
        if not self.schedule or self.is_climate_mode:
            return [
                mode.value
                for mode in WiserPresetOptionsEnum
                if mode != WiserPresetOptionsEnum.advance_schedule
            ]
        return [mode.value for mode in WiserPresetOptionsEnum]

    @property
    def away_mode_suppressed(self):
        """Get if away mode is suppressed for room"""
        return self._data.get("AwayModeSuppressed", TEXT_UNKNOWN)

    @property
    def boost_end_time(self) -> datetime:
        """Get boost end timestamp"""
        return datetime.fromtimestamp(self._data.get("OverrideTimeoutUnixTime", 0))

    @property
    def boost_time_remaining(self) -> int | float:
        """Get boost time remaining"""
        if self.is_boosted:
            return (self.boost_end_time - datetime.now()).total_seconds()
        else:
            return 0

    @property
    def current_control_source(self) -> str:
        """Get the current control source for the hot water"""
        return self._data.get("HotWaterDescription", TEXT_UNKNOWN)

    @property
    def current_state(self) -> str:
        """Get the current state of the hot water"""
        return self._data.get("WaterHeatingState", TEXT_UNKNOWN)

    @property
    def id(self) -> int | None:
        """Get the id of the hot water channel"""
        return self._data.get("id")

    @property
    def is_away_mode(self) -> bool:
        """Return if away mode is on."""
        return (
            True if self._data.get("HotWaterDescription") == "FromAwayMode" else False
        )

    @property
    def is_boosted(self) -> bool:
        """Get if the hot water is currently boosted"""
        return True if "Boost" in self._data.get("HotWaterDescription", None) else False

    @property
    def is_climate_mode(self) -> bool:
        """Return if hw climate mode set."""
        return self._wiser_rest_controller._api_parameters.hw_climate_mode

    @property
    def is_heating(self) -> bool:
        """Get if the hot water is currently heating"""
        return True if self._data.get("WaterHeatingState") == TEXT_ON else False

    @property
    def is_override(self) -> bool:
        """Get if hotwater has an override"""
        return (
            True
            if self._data.get("OverrideType", TEXT_UNKNOWN)
            not in [TEXT_UNKNOWN, "None"]
            else False
        )

    @property
    def is_climate_mode_off(self) -> bool:
        """Determine if hw is off for climate mode."""
        if self.is_climate_mode:
            if (
                (self._extra_config and self._extra_config.get("climate_off", False))
                and self._data.get("Mode") == WiserDeviceModeEnum.manual.value
                and not self.is_heating
            ):
                return True

            # If not off but extr_config says it is, update extra config.
            elif not self.is_boosted and self._extra_config:
                if self._extra_config.get("climate_off"):
                    asyncio.get_running_loop().create_task(
                        self._update_extra_config("climate_off", False)
                    )
        return False

    @property
    def manual_heat(self) -> bool:
        """Return if manual heat is active."""
        if self.is_climate_mode is False or self.is_climate_mode_off:
            return False
        elif self._extra_config:
            return self._extra_config.get("manual_heat")
        return self._default_extra_config.get("manual_heat")

    async def set_manual_heat(self, enabled: bool):
        """Set manual heat."""
        if self.is_climate_mode:
            await self._update_extra_config("manual_heat", enabled)

    @property
    def mode(self) -> str | None:
        """Get or set the current hot water mode (On, Off or Auto)"""
        try:
            return (
                TEXT_OFF
                if self.is_climate_mode_off and not self.is_boosted
                else self._data.get("Mode")
            )
        except KeyError:
            return None

    async def set_mode(
        self, mode: WiserDeviceModeEnum | WiserHotWaterClimateModeEnum | str
    ):
        """Set the hotwater mode."""
        if isinstance(mode, WiserDeviceModeEnum | WiserHotWaterClimateModeEnum):
            mode = mode.value
        if mode == WiserHotWaterClimateModeEnum.off.value:
            if self.is_climate_mode:
                await self._update_extra_config("climate_off", True)
                await self._send_command(
                    {
                        "Mode": WiserDeviceModeEnum.manual.value,
                        "ManualWaterHeatingState": TEXT_OFF,
                        "RequestOverride": {"Type": "None"},
                    }
                )
                return True
            else:
                raise ValueError(
                    f"{mode} is not a valid Hot Water mode.  Valid modes are {self.available_modes}"
                )
        elif is_value_in_list(mode, self.available_modes):
            if self.is_climate_mode:
                await self._update_extra_config("climate_off", False)
            result = await self._send_command(
                {"Mode": mode.title(), "RequestOverride": {"Type": "None"}}
            )
            return result
        else:
            raise ValueError(
                f"{mode} is not a valid Hot Water mode.  Valid modes are {self.available_modes}"
            )

    @property
    def name(self) -> str:
        return "HotWater"

    @property
    def preset_mode(self) -> str | None:
        """Get the current preset mode"""
        if self.is_boosted:
            return TEXT_BOOST
        else:
            return None

    async def set_preset(self, preset: WiserPresetOptionsEnum | str):
        """Set the preset mode"""
        if isinstance(preset, WiserPresetOptionsEnum):
            preset = preset.value

        # Is it valid preset option?
        if preset in self.available_presets:
            if preset == WiserPresetOptionsEnum.cancel_overrides.value:
                await self.cancel_overrides()
            elif preset == WiserPresetOptionsEnum.advance_schedule.value:
                await self.schedule_advance()
            elif preset.lower().startswith(TEXT_BOOST.lower()):
                # Lookup boost duration
                duration = WISER_BOOST_DURATION[preset]
                _LOGGER.debug("Boosting hotwater for %s mins", duration)
                await self.override_state_for_duration("On", duration)
        else:
            raise ValueError(
                f"{preset} is not a valid preset.  Valid presets are {self.available_presets}"
            )

    @property
    def product_type(self) -> str:
        return "HotWater"

    @property
    def schedule(self) -> _WiserSchedule:
        """Get the hot water schedule"""
        return self._schedule

    @property
    def schedule_id(self):
        """Get the hot water schedule id"""
        return self._data.get("ScheduleId", 0)

    @property
    def setpoint_origin(self) -> str:
        """Get the origin of the current status for the hotwater"""
        return self._data.get("HotWaterDescription", TEXT_UNKNOWN)

    @property
    def current_temperature(self) -> float:
        """Return current temperature."""
        return self._current_temperature

    @property
    def current_target_temperature(self) -> float:
        """Return current saved target temperature if in climate mode"""
        if self.is_climate_mode:
            if self._extra_config:
                return self._extra_config.get(
                    "target_temperature",
                    self._default_extra_config.get("target_temperature"),
                )
            return self._default_extra_config.get("target_temperature")

        return None

    @property
    def current_target_temperature_low(self) -> float:
        """Return current saved target temperature if in climate mode"""
        if self.is_climate_mode:
            if self._extra_config:
                return self._extra_config.get(
                    "target_temperature_low",
                    self._default_extra_config.get("target_temperature_low"),
                )
            return self._default_extra_config.get("target_temperature_low")

        return None

    @property
    def current_target_temperature_high(self) -> float:
        """Return current saved target temperature if in climate mode"""
        if self.is_climate_mode:
            if self._extra_config:
                return self._extra_config.get(
                    "target_temperature_high",
                    self._default_extra_config.get("target_temperature_high"),
                )
            return self._default_extra_config.get("target_temperature_high")

        return None

    async def set_target_temperature(self, temp: float) -> bool:
        """Set the target temperature of the hotwater to support climate mode."""
        if self.is_climate_mode:
            await self._update_extra_config("target_temperature", temp)
            return True
        raise WiserExtraConfigError(
            "You cannot set target temperature for hot water unless climate mode is not enabled."
        )

    async def set_target_temperature_low(self, temp):
        if not await self._update_extra_config("target_temperature_low", temp):
            _LOGGER.error(
                "Unable to set hotwater target lower temp.  This maybe caused by an issue with your extra config file."
            )
        await self._get_extra_config()

    async def set_target_temperature_high(self, temp):
        if not await self._update_extra_config("target_temperature_high", temp):
            _LOGGER.error(
                "Unable to set hotwater target upper temp.  This maybe caused by an issue with your extra config file."
            )
        await self._get_extra_config()

    async def boost(self, duration: int, state: str = TEXT_ON) -> bool:
        """
        Turn the hot water on for x minutes, overriding the current schedule or manual setting
        param duration: the duration to turn on for in minutes
        return: boolean
        """
        return await self.override_state_for_duration(state, duration)

    async def cancel_boost(self) -> bool | None:
        """
        Cancel the target temperature boost of the room
        return: boolean
        """
        if self.is_boosted:
            return await self.cancel_overrides()
        else:
            return True

    async def override_boost(self, state: str):
        """Override boost on/off to off/on but keep same remaining duration."""
        if (state == TEXT_ON and self.is_heating) or (
            state == TEXT_OFF and not self.is_heating
        ):
            return

        await self.override_state_for_duration(
            state, int(self.boost_time_remaining / 60)
        )

    async def cancel_overrides(self):
        """
        Cancel all overrides of the hot water
        return: boolean
        """
        if self.is_override:
            return await self._send_command({"RequestOverride": {"Type": "None"}})

    async def override_state(self, state: str) -> bool:
        """
        Override hotwater state.  In auto this is until the next scheduled event.  In manual mode this is until changed.
        return: boolean
        """
        if await self.cancel_boost():
            if state.casefold() == TEXT_ON.casefold():
                return await self._send_command(
                    {
                        "RequestOverride": {
                            "Type": "Manual",
                            "SetPoint": tf._to_wiser_temp(TEMP_HW_ON, "hotwater"),
                        }
                    }
                )
            elif state.casefold() == TEXT_OFF.casefold():
                return await self._send_command(
                    {
                        "RequestOverride": {
                            "Type": "Manual",
                            "SetPoint": tf._to_wiser_temp(TEMP_HW_OFF, "hotwater"),
                        }
                    }
                )
            else:
                raise ValueError(
                    f"Invalid state value {state}.  Should be {TEXT_ON} or {TEXT_OFF}"
                )
        return False

    async def override_state_for_duration(self, state: str, duration: int) -> bool:
        """
        Override the hot water state for x minutes, overriding the current schedule or manual setting
        param duration: the duration to turn on for in minutes
        return: boolean
        """
        if state.casefold() == TEXT_ON.casefold():
            return await self._send_command(
                {
                    "RequestOverride": {
                        "Type": "Manual",
                        "DurationMinutes": duration,
                        "SetPoint": tf._to_wiser_temp(TEMP_HW_ON, "hotwater"),
                    }
                }
            )
        elif state.casefold() == TEXT_OFF.casefold():
            return await self._send_command(
                {
                    "RequestOverride": {
                        "Type": "Manual",
                        "DurationMinutes": duration,
                        "SetPoint": tf._to_wiser_temp(TEMP_HW_OFF),
                    }
                }
            )
        else:
            raise ValueError(
                f"Invalid state value {state}.  Should be {TEXT_ON} or {TEXT_OFF}"
            )

    async def schedule_advance(self) -> bool:
        """
        Advance hot water schedule to the next scheduled state setting
        return: boolean
        """
        if self.schedule:
            if await self.cancel_boost():
                return await self.override_state(self.schedule.next.setting)
        return False



from ..const import TEXT_UNKNOWN


class _WiserSignalStrength(object):
    """Data structure for zigbee signal information for a Wiser device"""

    def __init__(self, data: dict):
        self._data = data

    @property
    def displayed_signal_strength(self) -> str:
        """Get the description of signal strength"""
        return self._data.get("DisplayedSignalStrength", TEXT_UNKNOWN)

    @property
    def controller_reception_lqi(self) -> int:
        """Get the signal lqi (quality) for the controller"""
        return self._data.get("ReceptionOfController", {"Lqi": 0}).get("Lqi", None)

    @property
    def controller_reception_rssi(self) -> int:
        """Get the rssi (strength) of the controller signal"""
        return self._data.get("ReceptionOfController", {"Rssi": 0}).get("Rssi", None)

    @property
    def controller_signal_strength(self) -> int:
        """Get the signal strength percent for the device"""
        return (
            min(100, int(2 * (self.controller_reception_rssi + 100)))
            if self.controller_reception_rssi and self.controller_reception_rssi != 0
            else 0
        )

    @property
    def device_reception_lqi(self) -> int:
        """Get the signal lqi (quality) for the device"""
        return self._data.get("ReceptionOfDevice", {"Lqi": 0}).get("Lqi", None)

    @property
    def device_reception_rssi(self) -> int:
        """Get the rssi (strength) of the device signal"""
        return self._data.get("ReceptionOfDevice", {"Rssi": 0}).get("Rssi", None)

    @property
    def device_signal_strength(self) -> int:
        """Get the signal strength percent for the device"""
        if self.device_reception_rssi:
            return (
                min(100, int(2 * (self.device_reception_rssi + 100)))
                if self.device_reception_rssi != 0
                else 0
            )
        return None



class _WiserGPS:
    """Data structure for gps positional information for a Wiser Hub"""

    def __init__(self, data: dict):
        self._data = data

    @property
    def latitude(self) -> float:
        """Get the latitude of the hub"""
        return self._data.get("Latitude")

    @property
    def longitude(self) -> float:
        """Get the longitude of the hub"""
        return self._data.get("Longitude")



import enum

from aioWiserHeatAPI.helpers.version import Version

# Temperature Constants
DEFAULT_AWAY_MODE_TEMP = 10.5
DEFAULT_DEGRADED_TEMP = 18
DEFAULT_BOOST_DELTA = 2
MAX_BOOST_INCREASE = 5
TEMP_ERROR = 2000
TEMP_MINIMUM = 5
TEMP_MAXIMUM = 30
TEMP_HW_ON = 110
TEMP_HW_OFF = -20
TEMP_OFF = -20


# Battery Constants
ROOMSTAT_MIN_BATTERY_LEVEL = 1.7
ROOMSTAT_FULL_BATTERY_LEVEL = 2.7
TRV_FULL_BATTERY_LEVEL = 3.0
TRV_BATTERY_LEVEL_MAPPING = {
    3.0: 100,
    2.9: 80,
    2.8: 60,
    2.7: 40,
    2.6: 20,
    2.5: 10,
    2.4: 5,
    2.3: 0,
}


# Other Constants
REST_RETRY_BACKOFF = [0.1, 0.5, 1, 3, 5]
REST_BACKOFF_FACTOR = 1
REST_RETRIES = 3
REST_TIMEOUT = 20

# Text Values
TEXT_AUTO = "Auto"
TEXT_BOOST = "Boost"
TEXT_CLOSE = "Close"
TEXT_DEGREESC = "DegreesC"
TEXT_HEATING = "Heating"
TEXT_LEVEL = "Level"
TEXT_LIGHTING = "Lighting"
TEXT_MANUAL = "Manual"
TEXT_NO_CHANGE = "NoChange"
TEXT_OFF = "Off"
TEXT_ON = "On"
TEXT_ONOFF = "OnOff"
TEXT_OPEN = "Open"
TEXT_PASSIVE = "Passive"
TEXT_SETPOINT = "Setpoint"
TEXT_SHUTTERS = "Shutters"
TEXT_STATE = "State"
TEXT_TEMP = "Temp"
TEXT_TIME = "Time"
TEXT_UNKNOWN = "Unknown"
TEXT_WEEKDAYS = "Weekdays"
TEXT_WEEKENDS = "Weekends"
TEXT_ALL = "All"
TEXT_UNABLE = "Unable"
# added by Lgo44

TEXT_REVERSE = "ReverseWithLoad"
TEXT_CONSISTENT = "ConsistentWithLoad"
TEXT_ALWAYSON = "AlwaysOn"
TEXT_ALWAYSOFF = "AlwaysOff"

TEXT_LASTON = "LastOnBrightness"
TEXT_LEVELPERCENT = "LevelPercent"

# End added by Lgo44

# Day Value Lists
WEEKDAYS = ["Monday", "Tuesday", "Wednesday", "Thursday", "Friday"]
WEEKENDS = ["Saturday", "Sunday"]
SPECIAL_DAYS = [TEXT_WEEKDAYS, TEXT_WEEKENDS, TEXT_ALL]
SPECIAL_TIMES = {"Sunrise": 3000, "Sunset": 4000}

OPENTHERMV2_MIN_VERSION = Version("4.32.47")
HUB_GEN2_MIN_HTTPS_VERSION = Version("4.42.22")

# Wiser Hub Rest Api URL Constants
WISERHUBURL = "{}:{}/data/v2/"
WISERHUBDOMAIN = WISERHUBURL + "domain/"
WISERHUBNETWORK = WISERHUBURL + "network/"
WISERHUBSCHEDULES = WISERHUBURL + "schedules/"
WISERHUBOPENTHERM = WISERHUBURL + "opentherm/"
WISERHUBOPENTHERMV2 = WISERHUBURL + "openTherm/"
WISERHUBSTATUS = WISERHUBURL + "status/"
WISERSYSTEM = "System"
WISERDEVICE = "Device/{}"
WISERHOTWATER = "HotWater/{}"
WISERROOM = "Room/{}"
WISERSMARTVALVE = "SmartValve/{}"
WISERROOMSTAT = "RoomStat/{}"
WISERSMARTPLUG = "SmartPlug/{}"
WISERHEATINGACTUATOR = "HeatingActuator/{}"
WISERUFHCONTROLLER = "UnderFloorHeating/{}"
WISERSHUTTER = "Shutter/{}"
WISERLIGHT = "Light/{}"
WISERPOWERTAGCONTROL = "PTC/{}"
WISERPOWERTAGENERGY = "PTE/{}"
WISERSMOKEALARM = "SmokeAlarmDevice/{}"
WISERBINARYSENSOR = "BinarySensor/{}"
WISERBOILERINTERFACE = "BoilerInterface/{}"
WISERBUTTONPANEL = "ButtonPanel/{}"
WISERTHRESHOLDSENSOR = "ThresholdSensor/{}"
WISERUICONFIGURATION = "UIConfiguration/{}"


# Enums
class WiserUnitsEnum(enum.Enum):
    imperial = "imperial"
    metric = "metric"


class WiserTempLimitsEnum(enum.Enum):
    heating = {"min": 5, "max": 30, "off": -20, "type": "range"}
    current = {"min": -19, "max": 99, "off": -20, "type": "range"}
    hotwater = {"on": 110, "off": -20, "type": "onoff"}
    boostDelta = {"min": 0, "max": 5, "type": "range"}
    floorHeatingMin = {"min": 5, "max": 39, "off": -20, "type": "range"}
    floorHeatingMax = {"min": 5, "max": 40, "off": -20, "type": "range"}
    floorHeatingOffset = {"min": -9, "max": 9, "type": "range"}


class WiserAwayActionEnum(enum.Enum):
    off = TEXT_OFF
    nochange = TEXT_NO_CHANGE


class WiserShutterAwayActionEnum(enum.Enum):
    close = TEXT_CLOSE
    nochange = TEXT_NO_CHANGE


# added by Lgo44
class WiserLightLedIndicatorEnum(enum.Enum):
    reverse = TEXT_REVERSE
    consistent = TEXT_CONSISTENT
    allwayson = TEXT_ALWAYSON
    allwaysoff = TEXT_ALWAYSOFF


class WiserLightPowerOnBehaviourEnum(enum.Enum):
    last = TEXT_LASTON
    levelpercent = TEXT_LEVELPERCENT


# End added by Lgo44


class WiserDeviceModeEnum(enum.Enum):
    auto = TEXT_AUTO
    manual = TEXT_MANUAL


class WiserHotWaterClimateModeEnum(enum.Enum):
    auto = TEXT_AUTO
    manual = TEXT_MANUAL
    off = TEXT_OFF


class WiserHeatingModeEnum(enum.Enum):
    off = TEXT_OFF
    auto = TEXT_AUTO
    manual = TEXT_MANUAL


class WiserPresetOptionsEnum(enum.Enum):
    advance_schedule = "Advance Schedule"
    cancel_overrides = "Cancel Overrides"
    boost30 = "Boost 30m"
    boost60 = "Boost 1h"
    boost120 = "Boost 2h"
    boost180 = "Boost 3h"


WISER_BOOST_DURATION = {
    "Boost 30m": 30,
    "Boost 1h": 60,
    "Boost 2h": 120,
    "Boost 3h": 180,
}


class WiserScheduleTypeEnum(enum.Enum):
    heating = TEXT_HEATING
    onoff = TEXT_ONOFF
    level = TEXT_LEVEL
    lighting = TEXT_LIGHTING
    shutters = TEXT_SHUTTERS


DEFAULT_LEVEL_SCHEDULE = {
    "Monday": {"Time": [], "Level": []},
    "Tuesday": {"Time": [], "Level": []},
    "Wednesday": {"Time": [], "Level": []},
    "Thursday": {"Time": [], "Level": []},
    "Friday": {"Time": [], "Level": []},
    "Saturday": {"Time": [], "Level": []},
    "Sunday": {"Time": [], "Level": []},
}



import inspect

from aioWiserHeatAPI import _LOGGER

from ..const import TEXT_UNKNOWN
from ..rest_controller import _WiserRestController


class _WiserUIConfigSensor:
    """Class representing a Wiser threshold sensor for a WIser device."""

    def __init__(
        self,
        wiser_rest_controller: _WiserRestController,
        endpoint: str,
        data: dict,
    ):
        self._wiser_rest_controller = wiser_rest_controller
        self._endpoint = endpoint
        self._data = data

    async def _send_command(self, cmd: dict):
        """
        Send control command to the UI Configuration
        param cmd: json command structure
        return: boolen - true = success, false = failed
        """
        result = await self._wiser_rest_controller._send_command(
            self._endpoint.format(self.id), cmd
        )
        if result:
            self._device_type_data = result
        if result:
            _LOGGER.debug(
                "Wiser light - {} command successful".format(
                    inspect.stack()[1].function
                )
            )
            return True
        return False

    @property
    def id(self) -> str:
        """Get the id of UIConfig sensor"""
        return self._data.get("id", 0)

    @property
    def brightness(self) -> str:
        """Get the brightness of ui config sensor"""
        return self._data.get("Brightness", 0)

    @property
    def inactive_brightness(self) -> str:
        """Get the inactive brightness of ui config sensor"""
        return self._data.get("InactiveBrightness", 0)

    @property
    def activity_timeout(self) -> str:
        """Get the activity timeout of ui config sensor"""
        return self._data.get("ActivityTimeout", 0)

    @property
    def uuid(self) -> str:
        """Get the UUID of ui config sensor"""
        return self._data.get("UUID", TEXT_UNKNOWN)

    async def set_brightness(self, enabled: bool):
        if await self._send_command({"Brightness": str(enabled).lower()}):
            return True

    async def set_inactive_brightness(self, enabled: bool):
        if await self._send_command({"InactiveBrightness": str(enabled).lower()}):
            return True



#!/usr/bin/env python3
"""
# Wiser API Version 2

angelosantagata@gmail.com
msparker@sky.com


https://github.com/asantaga/wiserheatingapi


This API allows you to get information from and control your wiserhub.
"""

import pathlib
from datetime import datetime
from typing import Any, Optional

from aioWiserHeatAPI.helpers.version import Version

from . import __VERSION__, _LOGGER
from .cli import log_response_to_file
from .const import (
    DEFAULT_AWAY_MODE_TEMP,
    DEFAULT_DEGRADED_TEMP,
    HUB_GEN2_MIN_HTTPS_VERSION,
    MAX_BOOST_INCREASE,
    OPENTHERMV2_MIN_VERSION,
    TEMP_ERROR,
    TEMP_HW_OFF,
    TEMP_HW_ON,
    TEMP_MAXIMUM,
    TEMP_MINIMUM,
    TEMP_OFF,
    WISERHUBDOMAIN,
    WISERHUBNETWORK,
    WISERHUBOPENTHERM,
    WISERHUBOPENTHERMV2,
    WISERHUBSCHEDULES,
    WISERHUBSTATUS,
    WISERHUBURL,
    WiserUnitsEnum,
)
from .devices import _WiserDeviceCollection
from .exceptions import (
    WiserHubAuthenticationError,
    WiserHubConnectionError,
    WiserHubRESTError,
    WiserScheduleError,
)
from .heating import _WiserHeatingChannelCollection
from .helpers.automations import _WiserHeatingChannelAutomations
from .helpers.status import WiserStatus
from .hot_water import _WiserHotwater
from .moments import _WiserMomentCollection
from .refactor import refactor
from .rest_controller import (
    WiserRestActionEnum,
    _WiserConnectionInfo,
    _WiserRestController,
)
from .room import _WiserRoomCollection
from .schedule import WiserScheduleTypeEnum, _WiserScheduleCollection
from .system import _WiserSystem


class WiserAPI:
    """
    Main api class to access all entities and attributes of wiser system
    """

    def __init__(
        self,
        host: str,
        secret: str,
        port: Optional[int] = 80,
        units: Optional[WiserUnitsEnum] = WiserUnitsEnum.metric,
        extra_config_file: Optional[str] = None,
        enable_automations: Optional[bool] = True,
    ):
        # Connection variables
        self._wiser_api_connection = _WiserConnectionInfo()
        self._wiser_rest_controller = None

        # Set connection params
        self._wiser_api_connection.host = host
        self._wiser_api_connection.secret = secret
        self._wiser_api_connection.port = port
        self._wiser_api_connection.units = units
        self._wiser_api_connection.extra_config_file = extra_config_file
        self._wiser_api_connection.enable_automations = enable_automations

        # Hub Data
        self._domain_data = {}
        self._network_data = {}
        self._schedule_data = {}
        self._opentherm_data = {}
        self._status_data = {}

        # Data stores for exposed properties
        self._devices = None
        self._hotwater = None
        self._heating_channels = None
        self._moments = None
        self._rooms = None
        self._schedules = None
        self._system = None

        self._enable_automations = enable_automations
        self._extra_config_file = extra_config_file
        self._extra_config = None

        # Log initialisation info
        _LOGGER.info(
            f"WiserHub API v{__VERSION__} Initialised - Host: {host}, Units: {self._wiser_api_connection.units.name.title()}, Extra Config: {self._extra_config_file}, Automations: {self._enable_automations}"
        )

        if (
            self._wiser_api_connection.host is not None
            and self._wiser_api_connection.secret is not None
        ):
            # Create an instance of the rest controller
            self._wiser_rest_controller = _WiserRestController(
                wiser_connection_info=self._wiser_api_connection
            )
        else:
            raise WiserHubConnectionError(
                "Missing or incomplete connection information"
            )

    async def read_hub_data(self):
        """Update data objects form the hub."""
        await self._build_objects()

        # Run automations
        if self._enable_automations:
            automations = _WiserHeatingChannelAutomations(
                self._wiser_rest_controller, self.heating_channels
            )
            if await automations.run_automations():
                await self._build_objects()

    async def get_hub_data(self) -> dict[str, Any]:
        """Get data from hub."""
        await self._get_hub_data()
        return self.raw_hub_data

    async def send_hub_command(
        self, path: str, payload: dict[str, Any]
    ) -> dict[str, Any]:
        """Send command to hub and return result."""
        if path and payload:
            response = await self._wiser_rest_controller._do_hub_action(
                WiserRestActionEnum.PATCH, WISERHUBURL + path, payload
            )
            return response

    async def _get_hub_data(self) -> bool:
        try:
            start_time = datetime.now()
            self._domain_data = await self._wiser_rest_controller.get_hub_data(
                WISERHUBDOMAIN
            )

            # Determine if we need to use https for v2 hubs.  FW needs to be 4.42.23 or higher
            if self._domain_data:
                hw_gen = self._domain_data.get("System", {}).get(
                    "HardwareGeneration", 1
                )
                fw_version = Version(
                    self._domain_data.get("System", {}).get(
                        "ActiveSystemVersion", "1.0.0"
                    )
                )
                _LOGGER.debug(
                    "Hub Hardware Generation: %s, Firmware Version: %s",
                    hw_gen,
                    fw_version,
                )
                if hw_gen == 2 and fw_version >= HUB_GEN2_MIN_HTTPS_VERSION:
                    _LOGGER.debug("Using HTTPS for Wiser Hub REST API calls")
                    self._wiser_rest_controller.use_https = True

            self._network_data = await self._wiser_rest_controller.get_hub_data(
                WISERHUBNETWORK
            )
            self._schedule_data = await self._wiser_rest_controller.get_hub_data(
                WISERHUBSCHEDULES
            )
            try:
                self._status_data = await self._wiser_rest_controller.get_hub_data(
                    WISERHUBSTATUS
                )
            except WiserHubRESTError:
                self._status_data = {}

            # Only get opentherm data if connected
            opentherm = self._domain_data.get("System", {}).get(
                "OpenThermConnectionStatus", ""
            )
            if opentherm and opentherm == "Connected" or True:
                fw_version = Version(
                    self._domain_data.get("System", {}).get(
                        "ActiveSystemVersion", "1.0.0"
                    )
                )
                hw_version = self._domain_data.get("System", {}).get(
                    "HardwareGeneration", 1
                )
                self._opentherm_data = await self._wiser_rest_controller.get_hub_data(
                    WISERHUBOPENTHERMV2
                    if hw_version == 2 and fw_version >= OPENTHERMV2_MIN_VERSION
                    else WISERHUBOPENTHERM,
                    False,
                )
        except (
            WiserHubConnectionError,
            WiserHubAuthenticationError,
            WiserHubRESTError,
            Exception,
        ) as ex:
            _LOGGER.debug("Update from Wiser hub failed. %s", ex)
            raise ex
        else:
            # Set hub name on rest controller
            self._wiser_rest_controller._hub_name = (
                self._network_data.get("Station", {})
                .get("NetworkInterface", {})
                .get("HostName", "")
            )

            _LOGGER.debug(
                "Update from %s successful and took %ss",
                self._wiser_rest_controller._hub_name,
                (datetime.now() - start_time).total_seconds(),
            )
            return True

    async def _build_objects(self):
        """Read all data from hub and populate objects"""

        # Read data from hub
        try:
            await self._get_hub_data()

            # load extra data
            self._wiser_rest_controller._extra_config_file = self._extra_config_file
            await self._wiser_rest_controller.get_extra_config_data()

            if self._domain_data != {} and self._network_data != {}:
                # System Object
                _device_data = self._domain_data.get("Device", [])
                self._system = _WiserSystem(
                    self._wiser_rest_controller,
                    self._domain_data,
                    self._network_data,
                    _device_data,
                    self._opentherm_data,
                )

                # Schedules Collection
                self._schedules = _WiserScheduleCollection(
                    self._wiser_rest_controller,
                    self._schedule_data,
                    self._system.sunrise_times,
                    self._system.sunset_times,
                )

                # Devices Collection
                self._devices = _WiserDeviceCollection(
                    self._wiser_rest_controller,
                    self._domain_data,
                    self._schedules,
                )

                # Rooms Collection
                room_data = self._domain_data.get("Room", [])
                self._rooms = _WiserRoomCollection(
                    self._wiser_rest_controller,
                    room_data,
                    self._schedules.get_by_type(WiserScheduleTypeEnum.heating),
                    self._devices,
                    self._enable_automations,
                )

                # Hot Water
                if self._domain_data.get("HotWater"):
                    schedule = self._schedules.get_by_id(
                        WiserScheduleTypeEnum.onoff,
                        self._domain_data.get("HotWater")[0].get("ScheduleId", 0),
                    )
                    self._hotwater = _WiserHotwater(
                        self._wiser_rest_controller,
                        self._domain_data.get("HotWater", {})[0],
                        schedule,
                    )

                # Heating Channels
                if self._domain_data.get("HeatingChannel"):
                    self._heating_channels = _WiserHeatingChannelCollection(
                        self._domain_data.get("HeatingChannel"), self._rooms
                    )

                # Moments
                if self._domain_data.get("Moment"):
                    self._moments = _WiserMomentCollection(
                        self._wiser_rest_controller,
                        self._domain_data.get("Moment"),
                    )

                # If gets here with no exceptions then success and return true
                return True
        except (
            WiserHubConnectionError,
            WiserHubAuthenticationError,
            WiserHubRESTError,
            Exception,
        ) as ex:
            raise ex

    # API properties
    @property
    def api_parameters(self):
        """Rest control api parameters."""
        return self._wiser_rest_controller._api_parameters

    @property
    def devices(self) -> _WiserDeviceCollection:
        """List of device entities attached to the Wiser Hub"""
        return self._devices

    @property
    def heating_channels(self) -> _WiserHeatingChannelCollection:
        """List of heating channel entities on the Wiser Hub"""
        return self._heating_channels

    @property
    def hotwater(self) -> _WiserHotwater:
        """List of hot water entities on the Wiser Hub"""
        return self._hotwater

    @property
    def moments(self) -> _WiserMomentCollection:
        """List of moment entities on the Wiser Hub"""
        return self._moments

    @property
    def rooms(self) -> _WiserRoomCollection:
        """List of room entities configured on the Wiser Hub"""
        return self._rooms

    @property
    def schedules(self) -> _WiserScheduleCollection:
        """List of schedules"""
        return self._schedules

    @property
    def status(self) -> WiserStatus:
        """Hub status info"""
        return WiserStatus(self._status_data)

    @property
    def system(self) -> _WiserSystem:
        """Entity of the Wiser Hub"""
        return self._system

    @property
    def units(self) -> WiserUnitsEnum:
        """Get or set units for temperature"""
        return self._wiser_api_connection.units

    @units.setter
    def units(self, units: WiserUnitsEnum):
        self._wiser_api_connection.units = units

    @property
    def version(self):
        """Return api version."""
        return __VERSION__

    @property
    def raw_hub_data(self):
        """Return raw hub data."""
        return {
            "Domain": self._domain_data,
            "Network": self._network_data,
            "Schedule": self._schedule_data,
            "OpenTherm": self._opentherm_data,
            "Status": self._status_data,
        }

    # @property
    # def refactored(self) -> dict[str, Any]:
    #    """Refactor json data."""
    #    return refactor(self.raw_hub_data)

    def output_raw_hub_data(
        self, data_class: str, filename: str, file_path: str
    ) -> bool:
        """Output raw hub data to json file"""
        # Get correct endpoint
        if data_class.lower() == "domain":
            endpoint = WISERHUBDOMAIN
        elif data_class.lower() == "network":
            endpoint = WISERHUBNETWORK
        elif data_class == "schedules":
            endpoint = WISERHUBSCHEDULES
        else:
            endpoint = None

        # Get raw json data
        if endpoint:
            data = self._wiser_rest_controller.get_hub_data(endpoint)
            try:
                if data:
                    # Write out to file
                    log_response_to_file(data, filename, False, pathlib.Path(file_path))
                    return True
            except Exception as ex:
                _LOGGER.error(ex)
                return False



"""
Handles power tag energy devices
"""

import inspect

from aioWiserHeatAPI import _LOGGER

from .const import TEXT_OFF, TEXT_ON, TEXT_UNABLE, TEXT_UNKNOWN, WISERDEVICE
from .helpers.device import _WiserElectricalDevice
from .helpers.equipment import _WiserEquipment


class _WiserPowerTagControl(_WiserElectricalDevice):
    """Class representing a Wiser Power Tag Energy device"""

    async def _send_command(self, cmd: dict, device_level: bool = False):
        """
        Send control command to the smart plug
        param cmd: json command structure
        return: boolen - true = success, false = failed
        """
        if device_level:
            result = await self._wiser_rest_controller._send_command(
                WISERDEVICE.format(self.id), cmd
            )
            if result:
                self._data = result
        else:
            result = await self._wiser_rest_controller._send_command(
                self._endpoint.format(self.device_type_id), cmd
            )
            if result:
                self._device_type_data = result
        if result:
            _LOGGER.debug(
                "Wiser light - {} command successful".format(
                    inspect.stack()[1].function
                )
            )
            return True
        return False

    @property
    def control_source(self) -> str:
        """Get the current control source."""
        return self._device_type_data.get("ControlSource", TEXT_UNKNOWN)

    @property
    def delivered_power(self) -> int:
        """Get the amount of current throught the plug over time"""
        return self._device_type_data.get("CurrentSummationDelivered", None)

    @property
    def equipment_id(self) -> int:
        """Get equipment id (v2 hub)"""
        return self._device_type_data.get("EquipmentId", 0)

    @property
    def equipment(self) -> _WiserEquipment | None:
        """Get equipment data"""
        return (
            _WiserEquipment(self._device_type_data.get("EquipmentData"))
            if self._device_type_data.get("EquipmentData")
            else None
        )

    @property
    def instantaneous_power(self) -> int:
        """Get the amount of current throught the plug now"""
        return self._device_type_data.get("InstantaneousDemand", None)

    @property
    def manual_state(self) -> str:
        """Get the current manual mode setting of the smart plug"""
        return self._device_type_data.get("ManualState", TEXT_UNKNOWN)

    @property
    def target_state(self) -> str:
        """Get the target state."""
        return self._device_type_data.get("TargetState", TEXT_UNKNOWN)

    @property
    def is_on(self) -> bool:
        """Get if the smart plug is on"""
        return True if self.output_state == TEXT_ON else False

    @property
    def output_state(self) -> str:
        """Get plug output state"""
        return self._device_type_data.get("OutputState", TEXT_OFF)

    @property
    def scheduled_state(self) -> str:
        """Get the current scheduled state of the smart plug"""
        return self._device_type_data.get("ScheduledState", TEXT_UNKNOWN)

    @property
    def actuator_type(self) -> str:
        """Get actuator type."""
        return self._device_type_data.get("ActuatorType", TEXT_UNKNOWN)

    @property
    def feedback_type(self) -> str:
        """Get feedback type."""
        return self._device_type_data.get("FeedbackType", TEXT_UNKNOWN)

    @property
    def polarity(self) -> str:
        """Get polarity."""
        return self._device_type_data.get("Polarity", TEXT_UNKNOWN)

    @property
    def relay_action(self) -> str:
        """Get relay action."""
        return self._device_type_data.get("RelayAction", TEXT_UNKNOWN)

    async def turn_on(self) -> bool:
        """
        Turn on the smart plug
        return: boolean
        """
        result = await self._send_command({"RequestOutput": TEXT_ON})
        if result:
            self._output_state = TEXT_ON
        return result

    async def turn_off(self) -> bool:
        """
        Turn off the smart plug
        return: boolean
        """
        result = await self._send_command({"RequestOutput": TEXT_OFF})
        if result:
            self._output_state = TEXT_OFF
        return result

    @property
    def energy_export(self) -> str:
        """Get energy export status"""
        return self._device_type_data.get("EnergyExport", TEXT_UNABLE)

    @property
    def rfid(self) -> int:
        """Get rfid of device"""
        return self._data.get("RfId", 0)

    @property
    def self_consumption(self) -> bool:
        """Get self consumption"""
        return self._device_type_data.get("SelfConsumption", False)


class _WiserPowerTagControlCollection:
    """Class representing collection of Power Tag Energy devices"""

    def __init__(self):
        self._items = []

    @property
    def all(self) -> list[_WiserPowerTagControl]:
        """Return all power tags"""
        return list(self._items)

    @property
    def count(self) -> int:
        """Return number of power tags"""
        return len(self.all)

    def get_by_id(self, device_id: int) -> _WiserPowerTagControl:
        """
        Gets a Power Tag Energy object from the device id
        param id: id of power tag
        return: _WiserPowerTagEnergy object
        """
        try:
            return [power_tag for power_tag in self.all if power_tag.id == device_id][0]
        except IndexError:
            return None

    def get_by_equipment_id(self, equipment_id: int) -> _WiserPowerTagControl:
        """
        Gets a Power Tag Energy object from the equipment id
        param id: id of power tag
        return: _WiserPowerTagEnergy object
        """
        try:
            return [
                power_tag for power_tag in self.all if power_tag.id == equipment_id
            ][0]
        except IndexError:
            return None



import hashlib
import inspect
from typing import Union
from uuid import UUID

from .. import _LOGGER
from ..const import (
    TEXT_UNKNOWN,
    WISERDEVICE,
    WiserAwayActionEnum,
    WiserDeviceModeEnum,
    WiserShutterAwayActionEnum,
)
from ..helpers.misc import is_value_in_list
from ..helpers.signal import _WiserSignalStrength
from ..rest_controller import _WiserRestController


class _WiserDevice(object):
    """Class representing a wiser heating device"""

    def __init__(
        self,
        wiser_rest_controller: _WiserRestController,
        endpoint: str,
        data: dict,
        device_type_data: dict,
        schedule: dict = None,
    ):
        self._away_action = None
        self._data = data
        self._device_type_data = device_type_data
        self._wiser_rest_controller = wiser_rest_controller
        self._endpoint = endpoint

    async def _send_command(self, cmd: dict, device_level: bool = False):
        """
        Send control command to the smart plug
        param cmd: json command structure
        return: boolen - true = success, false = failed
        """
        if device_level:
            result = await self._wiser_rest_controller._send_command(
                WISERDEVICE.format(self.id), cmd
            )
            if result:
                self._data = result
        else:
            result = await self._wiser_rest_controller._send_command(
                self._endpoint.format(self.id), cmd
            )
            if result:
                self._device_type_data = result
        if result:
            _LOGGER.debug(
                "Wiser device - {} command successful".format(
                    inspect.stack()[1].function
                )
            )
            return True
        return False

    @property
    def device_lock_enabled(self) -> bool:
        """Get or set device lock"""
        return self._data.get("DeviceLockEnabled", False)

    async def set_device_lock_enabled(self, enabled: bool = False):
        if await self._send_command({"DeviceLockEnabled": enabled}, True):
            self._device_lock_enabled = enabled
            return True

    @property
    def device_type(self) -> str:
        """Get the device types (hub v2 only)"""
        return self._data.get("Type", TEXT_UNKNOWN)

    @property
    def device_type_id(self) -> int:
        """Get the device id for the specific device type"""
        return self._device_type_data.get("id")

    @property
    def identify(self) -> bool:
        """Get or set if the identify function is enabled"""
        return self._data.get("IdentifyActive", False)

    async def set_identify(self, enabled: bool = False):
        if await self._send_command({"Identify": enabled}, True):
            self._indentify_active = enabled
            return True

    @property
    def firmware_version(self) -> str:
        """Get firmware version of device"""
        return self._data.get("ActiveFirmwareVersion", TEXT_UNKNOWN)

    @property
    def id(self) -> int:
        """Get id of device"""
        return self._data.get("id")

    @property
    def model(self) -> str:
        """Get model of device"""
        # Lights, shutters and power tags currently have model identifier as Unknowm
        # return (
        #    self._data.get("ProductType", TEXT_UNKNOWN)
        #    if self._data.get("ModelIdentifier") == TEXT_UNKNOWN
        #    else self._data.get("ModelIdentifier", TEXT_UNKNOWN)
        # )

        if self.product_model not in ["", TEXT_UNKNOWN]:
            return self._data.get("ProductModel")
        elif self._data.get("ModelIdentifier") != TEXT_UNKNOWN:
            return self._data.get("ModelIdentifier")
        else:
            return self.product_identifier

    @property
    def name(self) -> str:
        """Get name of device."""
        if self._data.get("Name"):
            return self._data.get("Name")
        return f"{self.product_type}-{self.id}"

    @property
    def node_id(self) -> int:
        """Get zigbee node id of device"""
        return self._data.get("NodeId", 0)

    @property
    def ota_hardware_version(self) -> int:
        """Get ota hardware version"""
        return self._data.get("OtaHardwareVersion", 0)

    @property
    def ota_version(self) -> int:
        """Get ota software version"""
        return self._data.get("OtaVersion", 0)

    @property
    def product_identifier(self) -> str:
        """Get product identifier of device"""
        return self._data.get("ProductIdentifier", TEXT_UNKNOWN)

    @property
    def product_model(self) -> str:
        """Get product model of device"""
        return self._data.get("ProductModel", TEXT_UNKNOWN)

    @property
    def parent_node_id(self) -> int:
        """Get zigbee node id of device this device is connected to"""
        return self._data.get("ParentNodeId", 0)

    @property
    def product_type(self) -> str:
        """Get product type of device"""
        return self._data.get("ProductType", TEXT_UNKNOWN)

    @property
    def room_id(self) -> int:
        """Get room id of the device."""
        return self._device_type_data.get("RoomId", self._data.get("RoomId", 0))

    @property
    def serial_number(self) -> str:
        """Get serial number of device"""
        return self._data.get("SerialNumber", TEXT_UNKNOWN)

    @property
    def signal(self) -> _WiserSignalStrength:
        """Get zwave network information"""
        return _WiserSignalStrength(self._data)

    # Added by LGO
    @property
    def endpoint(self) -> int:
        """Get endpoint"""
        return self._device_type_data.get("Endpoint", 0)

    # Device type Zigbee
    @property
    def type_comm(self) -> str:
        """Get type of zigbee device"""
        return self._data.get("Type", TEXT_UNKNOWN)

    # UUID Zigbee
    @property
    def uuid(self) -> str:
        """Get UUID zigbee"""
        if uuid := self._device_type_data.get("UUID"):
            return uuid
        uname = (
            f"{self.product_type}_{self.serial_number}_{self.id}_{self.device_type_id}"
        )
        hexstr = hashlib.md5(uname.encode("UTF-8")).hexdigest()
        return UUID(hex=hexstr)

    # END Added by LGO


class _WiserElectricalDevice(_WiserDevice):
    """Class representing a wiser electrical device"""

    def __init__(
        self,
        wiser_rest_controller: _WiserRestController,
        endpoint: str,
        data: dict,
        device_type_data: dict,
        schedule: dict,
    ):
        super().__init__(
            wiser_rest_controller, endpoint, data, device_type_data, schedule
        )
        self._schedule = schedule

        # Add device id to schedule
        if self._schedule:
            self._schedule._assignments.append(
                {"id": self.device_type_id, "name": self.name}
            )
            self._schedule._device_ids.append(self.id)

    @property
    def available_modes(self):
        """Get available modes"""
        return [mode.value for mode in WiserDeviceModeEnum]

    @property
    def available_away_mode_actions(self):
        """Get available away mode actions"""
        return [action.value for action in WiserAwayActionEnum]

    @property
    def away_mode_action(self) -> str:
        """Get or set the away action of the device (off or no change)"""
        return self._device_type_data.get("AwayAction", TEXT_UNKNOWN)

    async def set_away_mode_action(
        self,
        action: Union[WiserAwayActionEnum, WiserShutterAwayActionEnum, str],
    ) -> bool:
        if isinstance(action, WiserAwayActionEnum) or isinstance(
            action, WiserShutterAwayActionEnum
        ):
            action = action.value

        if is_value_in_list(action, self.available_away_mode_actions):
            if await self._send_command({"AwayAction": action}):
                self._away_action = action
                return True
        else:
            raise ValueError(
                f"{action} is not a valid away mode action.  "
                f"Valid modes are {self.available_away_mode_actions}"
            )

    @property
    def device_type_id(self) -> int:
        """Get the device id for the specific device type"""
        return self._device_type_data.get("id")

    @property
    def mode(self) -> str:
        """Get or set the current mode of the device (Manual or Auto)"""
        return self._device_type_data.get("Mode", TEXT_UNKNOWN)

    async def set_mode(self, mode: Union[WiserDeviceModeEnum, str]) -> bool:
        if isinstance(mode, WiserDeviceModeEnum):
            mode = mode.value
        if is_value_in_list(mode, self.available_modes):
            return await self._send_command({"Mode": mode.title()})
        else:
            raise ValueError(
                f"{mode} is not a valid mode.  Valid modes are {self.available_modes}"
            )

    @property
    def name(self) -> str:
        if self._data.get("Name"):
            return self._data.get("Name")
        return self._device_type_data.get("Name", TEXT_UNKNOWN)

    async def set_name(self, name: str) -> bool:
        if await self._send_command({"Name": name}):
            return True

    @property
    def schedule(self):
        """Get the schedule of the device"""
        return self._schedule

    @property
    def schedule_id(self) -> int:
        """Get the schedule id for the device"""
        return self._device_type_data.get("ScheduleId", 0)



from .helpers.battery import _WiserBattery
from .helpers.device import _WiserDevice
from .helpers.temp import _WiserTemperatureFunctions as tf


class _WiserSmokeAlarm(_WiserDevice):
    """Class representing a Wiser Smoke Alarm device"""

    @property
    def room_id(self) -> int:
        """Return room_id."""
        return self._data.get("RoomId")

    @property
    def alarm_sound_level(self) -> int:
        """Get the alarm sound level"""
        return self._device_type_data.get("AlarmSoundLevel")

    @property
    def alarm_sound_mode(self) -> int:
        """Get the alarm sound mode"""
        return self._device_type_data.get("AlarmSoundMode")

    @property
    def battery(self):
        """Get battery information for smoke alarm"""
        return _WiserBattery(self._data)

    @property
    def current_temperature(self) -> float:
        """Get the current temperature measured by the smoke alarm"""
        return tf._from_wiser_temp(
            self._device_type_data.get("MeasuredTemperature"), "current"
        )

    @property
    def led_brightness(self) -> int:
        """Get the led brightness"""
        return self._device_type_data.get("LEDBrightness")

    @property
    def life_time(self) -> int:
        """Get the life time"""
        return self._device_type_data.get("LifeTime")

    @property
    def hush_duration(self) -> int:
        """Get the hush duration"""
        return self._device_type_data.get("HushDuration")

    @property
    def smoke_alarm(self) -> bool:
        """Get if smoke alarm active"""
        return self._device_type_data.get("SmokeAlarm", False)

    @property
    def heat_alarm(self) -> bool:
        """Get if heat alarm active"""
        return self._device_type_data.get("HeatAlarm", False)

    @property
    def tamper_alarm(self) -> bool:
        """Get if tamper alarm active"""
        return self._device_type_data.get("Tamper", False)

    @property
    def fault_warning(self) -> bool:
        """Get if fault warning active"""
        return self._device_type_data.get("FaultWarning", False)

    @property
    def ac_mains(self) -> bool:
        """Get if on ac mains power"""
        return self._device_type_data.get("ACMains", False)

    @property
    def test_mode(self) -> bool:
        """Get if in test mode"""
        return self._device_type_data.get("TestMode", False)

    @property
    def battery_defect(self) -> bool:
        """Get if battery defect detected"""
        return self._device_type_data.get("BatteryDefect", False)

    @property
    def remote_alarm(self) -> bool:
        """Get if remote alarm active"""
        return self._device_type_data.get("RemoteAlarm", False)

    @property
    def hush_mode(self) -> bool:
        """Get if in hush mode"""
        return self._device_type_data.get("HushMode", False)

    @property
    def report_count(self) -> int:
        """Get the report count"""
        return self._device_type_data.get("ReportCount")

    @property
    def notification_enabled(self) -> bool:
        """Get if notifications active"""
        return self._device_type_data.get("EnableNotification", False)

    @property
    def supervision_notify(self) -> bool:
        """Get if supervision notify active"""
        return self._device_type_data.get("SupervisionNotify", False)

    @property
    def restore_notify(self) -> bool:
        """Get if restore notify active"""
        return self._device_type_data.get("RestoreNotify", False)


class _WiserSmokeAlarmCollection(object):
    """Class holding all wiser smoke alarms"""

    def __init__(self):
        self._items = []

    @property
    def all(self) -> list[_WiserSmokeAlarm]:
        return list(self._items)

    @property
    def count(self) -> int:
        return len(self.all)

    def get_by_id(self, smokealarm_id: int) -> _WiserSmokeAlarm:
        """
        Gets a smoke alarm object from the smoke alarms id
        param id: id of smoke alarm
        return: _WiserSmokeAlarm object
        """
        try:
            return [
                smokealarm for smokealarm in self.all if smokealarm.id == smokealarm_id
            ][0]
        except IndexError:
            return None



from .helpers.device import _WiserDevice


class _WiserBoilerInterface(_WiserDevice):
    """Class representing a Wiser Boiler Interface"""

    @property
    def heating_channel_ids(self) -> list[int]:
        """Heating channel ids."""
        return self._device_type_data.get("HeatingChannelIds")


class _WiserBoilerInterfaceCollection:
    """Class holding all Wiser Boiler Interfaces"""

    def __init__(self):
        self._items = []

    @property
    def all(self) -> list[_WiserBoilerInterface]:
        return list(self._items)

    @property
    def count(self) -> int:
        return len(self.all)

    def get_by_id(self, id: int) -> _WiserBoilerInterface:
        """
        Gets a boiler interface object from the binary sensor id
        param id: id of boiler interface
        return: _WiserBoilerInterface object
        """
        try:
            return [
                boilerinterface
                for boilerinterface in self.all
                if boilerinterface.id == id
            ][0]
        except IndexError:
            return None



from .helpers.battery import _WiserBattery
from .helpers.device import _WiserDevice
from .helpers.temp import _WiserTemperatureFunctions as tf


class _WiserSmartValve(_WiserDevice):
    """Class representing a Wiser Smart Valve device"""

    @property
    def battery(self):
        """Get battery information for smart valve"""
        return _WiserBattery(self._data)

    @property
    def current_target_temperature(self) -> float:
        """Get the smart valve current target temperature setting"""
        return tf._from_wiser_temp(self._device_type_data.get("SetPoint"))

    @property
    def current_temperature(self) -> float:
        """Get the current temperature measured by the smart valve"""
        return tf._from_wiser_temp(
            self._device_type_data.get("MeasuredTemperature"), "current"
        )

    @property
    def mounting_orientation(self) -> str:
        """Get the mouting orientation of the smart valve"""
        return self._device_type_data.get("MountingOrientation")

    @property
    def percentage_demand(self) -> int:
        """Get the current percentage demand of the smart valve"""
        return self._device_type_data.get("PercentageDemand")


class _WiserSmartValveCollection(object):
    """Class holding all wiser smart valves"""

    def __init__(self):
        self._items = []

    @property
    def all(self) -> list[_WiserSmartValve]:
        return list(self._items)

    @property
    def count(self) -> int:
        return len(self.all)

    def get_by_id(self, smartvalve_id: int) -> _WiserSmartValve:
        """
        Gets a SmartValve object from the SmartValves id
        param id: id of smart valve
        return: _WiserSmartValve object
        """
        try:
            return [
                smartvalve
                for smartvalve in self.all
                if smartvalve.id == smartvalve_id
            ][0]
        except IndexError:
            return None



import logging

from ..const import WiserHeatingModeEnum
from ..heating import _WiserHeatingChannelCollection
from ..hot_water import _WiserHotwater
from ..rest_controller import _WiserRestController

_LOGGER = logging.getLogger(__name__)


class _WiserHotWaterAutomations:
    def __init__(self, rest_controller: _WiserRestController, hotwater: _WiserHotwater):
        self._wiser_rest_controller = rest_controller
        self._hotwater = hotwater

        self._target_temperature: int = 0
        self._current_temperature: int = 0

    async def run_automations(self) -> bool:
        return await self.hotwater_control()

    @property
    def current_temperature(self) -> int:
        """Return current temperature"""
        return self._current_temperature

    @current_temperature.setter
    def current_temperature(self, temp: int):
        """Set current temperature."""
        self._current_temperature = temp

    @property
    def target_temperature(self) -> int:
        """Return target temperature"""
        return self._target_temperature

    @target_temperature.setter
    def target_temperature(self, temp: int):
        """Set target temperature."""
        self._target_temperature = temp

    async def hotwater_control(self) -> bool:
        """Control hotwater."""
        return True


class _WiserHeatingChannelAutomations:
    def __init__(
        self,
        rest_controller: _WiserRestController,
        heating_channels: _WiserHeatingChannelCollection,
    ):
        self._wiser_rest_controller = rest_controller
        self._heating_channels = heating_channels.all
        self._rooms = heating_channels._rooms.all

    async def run_automations(self) -> bool:
        return await self.passive_mode_control()

    async def passive_mode_control(self) -> bool:
        hub_updated: bool = False
        passive_mode_increment = (
            self._wiser_rest_controller._api_parameters.passive_mode_increment
        )
        # iterate each heating channel
        for heating_channel in self._heating_channels:
            passive_rooms = [
                room
                for room in self._rooms
                if room.is_passive_mode
                and room.id in heating_channel.room_ids
                and room.current_temperature  # added to prevent error if trv offline
            ]

            # If no passive mode rooms
            if not passive_rooms:
                _LOGGER.debug("No passive rooms exist")
                continue

            active_heating_rooms = [
                room
                for room in self._rooms
                if room.id in heating_channel.room_ids
                and (not room.is_passive_mode)
                and room.percentage_demand > 0
            ]

            _LOGGER.debug(
                f"Heating Channel {heating_channel.id}, Passive Rooms {[room.name for room in passive_rooms]}, Active Rooms {[room.name for room in active_heating_rooms]}"
            )

            # If any active rooms are heating
            if active_heating_rooms:
                for room in passive_rooms:
                    # If room is boosted do not override
                    if not room.is_boosted:
                        # Set target temp to heat passive room in
                        # increments of passive_temperature_increment (default 0.5)
                        target_temp = min(
                            round(
                                (room.current_temperature + passive_mode_increment) * 2
                            )
                            / 2,
                            (
                                room.schedule.current_setting
                                if room.mode == WiserHeatingModeEnum.auto.value
                                and room.schedule
                                else room.passive_mode_upper_temp
                            ),
                        )

                        if target_temp < room.passive_mode_lower_temp:
                            target_temp = room.passive_mode_lower_temp

                        if target_temp != room.current_target_temperature:
                            _LOGGER.debug(
                                f"Setting {room.name} to {target_temp}C caused by active rooms on heating channel {heating_channel.id}"
                            )
                            await room.set_target_temperature(target_temp)
                            hub_updated = True
            else:
                # Stop any passive rooms heating by setting to min temp
                for room in passive_rooms:
                    if (
                        room.current_target_temperature != room.passive_mode_lower_temp
                    ) and not room.is_boosted:
                        _LOGGER.debug(
                            f"Setting {room.name} to {room.passive_mode_lower_temp}C caused by no active rooms on heating channel {heating_channel.id}"
                        )
                        await room.set_target_temperature(room.passive_mode_lower_temp)
                        hub_updated = True

        return hub_updated



"""
Module to manage all devices
"""

from dataclasses import dataclass
from typing import Any

from .binary_sensor import (
    _WiserBinarySensor,
    _WiserBinarySensorCollection,
)
from .boiler_interface import _WiserBoilerInterface, _WiserBoilerInterfaceCollection
from .button_panel import _WiserButtonPanel, _WiserButtonPanelCollection
from .const import (
    WISERBINARYSENSOR,
    WISERBOILERINTERFACE,
    WISERBUTTONPANEL,
    WISERHEATINGACTUATOR,
    WISERLIGHT,
    WISERPOWERTAGCONTROL,
    WISERPOWERTAGENERGY,
    WISERROOMSTAT,
    WISERSHUTTER,
    WISERSMARTPLUG,
    WISERSMARTVALVE,
    WISERSMOKEALARM,
    WISERTHRESHOLDSENSOR,
    WISERUFHCONTROLLER,
    WISERUICONFIGURATION,
)
from .heating_actuator import _WiserHeatingActuator, _WiserHeatingActuatorCollection
from .helpers.device import _WiserDevice
from .helpers.threshold import (
    _WiserThresholdSensor,
)
from .helpers.uiconfiguration import _WiserUIConfigSensor
from .light import _WiserDimmableLight, _WiserLight, _WiserLightCollection
from .ptc import _WiserPowerTagControl, _WiserPowerTagControlCollection
from .pte import _WiserPowerTagEnergy, _WiserPowerTagEnergyCollection
from .rest_controller import _WiserRestController
from .roomstat import _WiserRoomStat, _WiserRoomStatCollection
from .schedule import WiserScheduleTypeEnum, _WiserScheduleCollection
from .shutter import _WiserShutter, _WiserShutterCollection
from .smartplug import _WiserSmartPlug, _WiserSmartPlugCollection
from .smartvalve import _WiserSmartValve, _WiserSmartValveCollection
from .smokealarm import _WiserSmokeAlarm, _WiserSmokeAlarmCollection
from .temp_humidity import _WiserTempHumidity, _WiserTempHumidityCollection
from .ufh import _WiserUFHController, _WiserUFHControllerCollection


@dataclass(frozen=True, kw_only=True)
class DeviceConfig:
    device_class: Any
    collection: Any
    endpoint: str = None
    heating: bool = False
    schedule_type: WiserScheduleTypeEnum | None = None
    has_v2_equipment: bool = False


@dataclass(frozen=True, kw_only=True)
class AncillaryDeviceConfig:
    device_class: Any
    attribute: str
    endpoint: str = None


PRODUCT_TYPE_CONFIG = {
    "SmartValve": DeviceConfig(
        device_class=_WiserSmartValve,
        collection=_WiserSmartValveCollection,
        endpoint=WISERSMARTVALVE,
        heating=True,
    ),
    "RoomStat": DeviceConfig(
        device_class=_WiserRoomStat,
        collection=_WiserRoomStatCollection,
        endpoint=WISERROOMSTAT,
        heating=True,
    ),
    "SmartPlug": DeviceConfig(
        device_class=_WiserSmartPlug,
        collection=_WiserSmartPlugCollection,
        endpoint=WISERSMARTPLUG,
        schedule_type=WiserScheduleTypeEnum.onoff,
        has_v2_equipment=True,
    ),
    "HeatingActuator": DeviceConfig(
        device_class=_WiserHeatingActuator,
        collection=_WiserHeatingActuatorCollection,
        endpoint=WISERHEATINGACTUATOR,
        heating=True,
        has_v2_equipment=True,
    ),
    "UnderFloorHeating": DeviceConfig(
        device_class=_WiserUFHController,
        collection=_WiserUFHControllerCollection,
        endpoint=WISERUFHCONTROLLER,
        heating=True,
    ),
    "Light": DeviceConfig(
        device_class=_WiserLight,
        collection=_WiserLightCollection,
        endpoint=WISERLIGHT,
        schedule_type=WiserScheduleTypeEnum.level,
    ),
    "Shutter": DeviceConfig(
        device_class=_WiserShutter,
        collection=_WiserShutterCollection,
        endpoint=WISERSHUTTER,
        schedule_type=WiserScheduleTypeEnum.level,
    ),
    "PTC": DeviceConfig(
        device_class=_WiserPowerTagControl,
        collection=_WiserPowerTagControlCollection,
        endpoint=WISERPOWERTAGCONTROL,
        has_v2_equipment=True,
    ),
    "PTE": DeviceConfig(
        device_class=_WiserPowerTagEnergy,
        collection=_WiserPowerTagEnergyCollection,
        endpoint=WISERPOWERTAGENERGY,
        has_v2_equipment=True,
    ),
    "SmokeAlarmDevice": DeviceConfig(
        device_class=_WiserSmokeAlarm,
        collection=_WiserSmokeAlarmCollection,
        endpoint=WISERSMOKEALARM,
    ),
    "BinarySensor": DeviceConfig(
        device_class=_WiserBinarySensor,
        collection=_WiserBinarySensorCollection,
        endpoint=WISERBINARYSENSOR,
    ),
    "BoilerInterface": DeviceConfig(
        device_class=_WiserBoilerInterface,
        collection=_WiserBoilerInterfaceCollection,
        endpoint=WISERBOILERINTERFACE,
    ),
    "ButtonPanel": DeviceConfig(
        device_class=_WiserButtonPanel,
        collection=_WiserButtonPanelCollection,
        endpoint=WISERBUTTONPANEL,
    ),
    "TempHumidity": DeviceConfig(
        device_class=_WiserTempHumidity,
        collection=_WiserTempHumidityCollection,
        endpoint=None,
    ),
}

ANCILLARY_SENSOR_CONFIG = {
    "ThresholdSensor": AncillaryDeviceConfig(
        device_class=_WiserThresholdSensor,
        attribute="_threshold_sensors",
        endpoint=WISERTHRESHOLDSENSOR,
    ),
    "UIConfiguration": AncillaryDeviceConfig(
        device_class=_WiserUIConfigSensor,
        attribute="_uiconfig_sensors",
        endpoint=WISERUICONFIGURATION,
    ),
}


class _WiserDeviceCollection:
    """Class holding all wiser devices"""

    def __init__(
        self,
        wiser_rest_controller: _WiserRestController,
        domain_data: dict,
        schedules: _WiserScheduleCollection,
    ):
        self._wiser_rest_controller = wiser_rest_controller
        self._device_data = domain_data.get("Device", {})
        self._domain_data = domain_data
        self._equipment_data = domain_data.get("Equipment", {})  # Supported by v2 hub
        self._schedules = schedules
        self._device_collection = {}

        self._build()

    def _get_equipment_data(self, equipment_id: int) -> dict:
        """Get equipment data"""
        equipment_data = [
            equipment_data
            for equipment_data in self._equipment_data
            if equipment_data.get("id") == equipment_id
        ]
        return equipment_data[0] if equipment_data else None

    def _get_device_info_by_id(self, device_id: int) -> dict[str, Any] | None:
        """Get device entry by id."""
        for device in self._domain_data.get("Device"):
            if device.get("id") == device_id:
                return device
        return None

    def _build(self):
        """Updated builf collection of devices.

        Starts with device type key first and then gets matching device data
        """
        # TODO - limit checks based on DeviceCapabilityMatrix
        for device_type in PRODUCT_TYPE_CONFIG:
            # If not yet in the collection, add empty collection to device collections.
            if device_type not in self._device_collection:
                self._device_collection[device_type] = PRODUCT_TYPE_CONFIG[
                    device_type
                ].collection()

            if device_type == "TempHumidity":
                # This is a workaround for temp humidity device(s)
                for device in self._domain_data.get("Device"):
                    if device.get("ProductType") == "TemperatureHumiditySensor":
                        device_config = PRODUCT_TYPE_CONFIG[device_type]
                        device_class = device_config.device_class
                        self._device_collection[device_type]._items.append(
                            device_class(
                                self._wiser_rest_controller,
                                device_config.endpoint,
                                device,
                                {},
                                None,
                            )
                        )

            elif device_type_data := self._domain_data.get(device_type):
                # We have this device type in data
                device_config = PRODUCT_TYPE_CONFIG[device_type]

                # Now build collection of devices
                for device in device_type_data:
                    # Get class to create and collection to append for device
                    device_class = device_config.device_class

                    # TODO: Fix this so not fudged
                    if device_type == "Light" and device.get("IsDimmable"):
                        device_class = _WiserDimmableLight

                    # Device info id can be DeviceId or id depending on device type
                    device_info_id = device.get("DeviceId", device.get("id"))

                    # Get device info data
                    device_info_data = self._get_device_info_by_id(device_info_id)

                    # If heating device add room id
                    if device_config.heating:
                        if not device.get("RoomId", device_info_data.get("RoomId")):
                            device["RoomId"] = self._get_temp_device_room_id(
                                self._domain_data, device_info_id
                            )

                    # If schedule device add schedule
                    if device_config.schedule_type:
                        device_schedule = self._schedules.get_by_id(
                            device_config.schedule_type, device.get("ScheduleId")
                        )
                    else:
                        device_schedule = None

                    # If has equipment data add to device info
                    if equipment_id := device.get("EquipmentId"):
                        device["EquipmentData"] = self._get_equipment_data(equipment_id)

                    # Add device to collection
                    self._device_collection[device_type]._items.append(
                        device_class(
                            self._wiser_rest_controller,
                            device_config.endpoint,
                            device_info_data,
                            device,
                            device_schedule,
                        )
                    )

        # Add ancillary sensors to device
        for anc_device_type, anc_device_type_info in ANCILLARY_SENSOR_CONFIG.items():
            if anc_device_type_data := self._domain_data.get(anc_device_type):
                for anc_device in anc_device_type_data:
                    device_id = anc_device.get("DeviceId", anc_device.get("id"))
                    if device := self.get_by_id(device_id):
                        if hasattr(device, anc_device_type_info.attribute):
                            getattr(device, anc_device_type_info.attribute).append(
                                anc_device_type_info.device_class(
                                    self._wiser_rest_controller,
                                    anc_device_type_info.endpoint,
                                    anc_device,
                                )
                            )

    def _get_temp_device_room_id(self, domain_data: dict, device_id: int) -> int:
        rooms = domain_data.get("Room")
        for room in rooms:
            room_device_list = []
            room_device_list.extend(room.get("SmartValveIds", []))
            room_device_list.extend(room.get("HeatingActuatorIds", []))
            room_device_list.append(room.get("RoomStatId"))
            room_device_list.append(room.get("UnderFloorHeatingId"))
            if device_id in room_device_list:
                return room.get("id")
        return 0

    @property
    def all(self) -> list[_WiserDevice]:
        """Return all devices"""
        items = []
        for key in self._device_collection:
            items.extend(self._device_collection[key].all)
        return items

    @property
    def count(self) -> int:
        """Return count of devices"""
        return len(self.all)

    @property
    def heating_actuators(self) -> _WiserHeatingActuatorCollection:
        """Return all heating actuators"""
        try:
            return self._device_collection["HeatingActuator"]
        except KeyError:
            return None

    @property
    def lights(self) -> _WiserLightCollection:
        """Return all lights"""
        try:
            return self._device_collection["Light"]
        except KeyError:
            return None

    @property
    def power_tags_c(self) -> _WiserPowerTagControlCollection:
        """Return all power tag Cs"""
        try:
            return self._device_collection["PTC"]
        except KeyError:
            return None

    @property
    def power_tags(self) -> _WiserPowerTagEnergyCollection:
        """Return all power tags"""
        try:
            return self._device_collection["PTE"]
        except KeyError:
            return None

    @property
    def roomstats(self) -> _WiserRoomStatCollection:
        """Return all roomstats"""
        try:
            return self._device_collection["RoomStat"]
        except KeyError:
            return None

    @property
    def shutters(self) -> _WiserShutterCollection:
        """Return all shutters"""
        try:
            return self._device_collection["Shutter"]
        except KeyError:
            return None

    @property
    def smartplugs(self) -> _WiserSmartPlugCollection:
        """Return all smart plugs"""
        try:
            return self._device_collection["SmartPlug"]
        except KeyError:
            return None

    @property
    def smartvalves(self) -> _WiserSmartValveCollection:
        """Return all smart valves (iTRVs)"""
        try:
            return self._device_collection["SmartValve"]
        except KeyError:
            return None

    @property
    def smokealarms(self) -> _WiserSmokeAlarmCollection:
        """Return all smoke alarms"""
        try:
            return self._device_collection["SmokeAlarmDevice"]
        except KeyError:
            return None

    @property
    def ufh_controllers(self) -> _WiserUFHControllerCollection:
        """Return all UFH controllers"""
        try:
            return self._device_collection["UnderFloorHeating"]
        except KeyError:
            return None

    @property
    def binary_sensor(self):
        """Return all binary sensors"""
        try:
            return self._device_collection["BinarySensor"]
        except KeyError:
            return None

    @property
    def temp_humidity_sensors(self):
        """Return all temp humidity sensors"""
        try:
            return self._device_collection["TempHumidity"]
        except KeyError:
            return None

    def get_by_id(self, device_id: int) -> _WiserDevice:
        """
        Gets a device object from the devices id
        param id: id of device
        return: Device type class
        """
        try:
            return [device for device in self.all if device.id == device_id][0]
        except IndexError:
            return None

    def get_by_room_id(self, room_id: int) -> list[_WiserDevice]:
        """
        Gets a list of devices belonging to the room id
        param room_id: the id of the room
        return: Device type class
        """
        try:
            return [device for device in self.all if device.room_id == room_id]
        except IndexError:
            return None

    def get_by_node_id(self, node_id: int) -> _WiserDevice:
        """
        Gets a device object from the devices zigbee node id
        param node_id: zigbee node id of device
        return: Device type class
        """
        try:
            return [device for device in self.all if device.node_id == node_id][0]
        except IndexError:
            return None

    def get_by_serial_number(self, serial_number: str) -> _WiserDevice:
        """
        Gets a device object from the devices serial number
        param node_id: serial number of device
        return: Device type class
        """
        try:
            return [
                device for device in self.all if device.serial_number == serial_number
            ][0]
        except IndexError:
            return None

    def get_by_parent_node_id(self, node_id: int) -> list[_WiserDevice]:
        """
        Gets a list of device from the devices zigbee parent node id
        param node_id: zigbee parent node id of device
        return: Device type class
        """
        try:
            return [device for device in self.all if device.parent_node_id == node_id]
        except IndexError:
            return None



from ..const import TEXT_UNKNOWN


class _WiserZigbee:
    """Data structure for zigbee information for a Wiser Hub"""

    def __init__(self, data: dict):
        self._data = data

    @property
    def error_72_reset(self) -> int:
        """Get error72reset info"""
        return self._data.get("Error72Reset", 0)

    @property
    def jpan_count(self) -> int:
        """Get jpan count info"""
        return self._data.get("JPANCount", 0)

    @property
    def network_channel(self) -> int:
        """Get network channel info"""
        return self._data.get("NetworkChannel", 0)

    @property
    def no_signal_reset(self) -> int:
        """Get no signal reset info"""
        return self._data.get("NoSignalReset", 0)

    @property
    def module_version(self) -> str:
        """Get zigbee module version info"""
        return self._data.get("ZigbeeModuleVersion", TEXT_UNKNOWN)

    @property
    def eui(self) -> str:
        """Get zigbee eui info"""
        return self._data.get("ZigbeeEUI", TEXT_UNKNOWN)



from ..rest_controller import _WiserRestController
from ..const import TEXT_UNKNOWN, WISERHUBNETWORK


class _WiserDetectedNetwork:
    """Data structure for detected network"""

    def __init__(self, data: dict):
        self._data = data

    @property
    def ssid(self) -> str:
        return self._data.get("SSID")

    @property
    def channel(self) -> int:
        return self._data.get("Channel")

    @property
    def security_mode(self) -> str:
        return self._data.get("SecurityMode")

    @property
    def rssi(self) -> int:
        return self._data.get("RSSI")


class _WiserNetworkStatistics:
    """Data structure for network statistics"""

    def __init__(self, data: dict):
        self._data = data

    @property
    def ipv4_address_ping(self) -> str:
        return self._data.get("IPv4AddressPing")

    @property
    def healthcheck_enabled(self) -> str:
        return self._data.get("HealthCheckingEnabled")

    @property
    def external_ping_enabled(self) -> str:
        return self._data.get("ExternalPingEnabled")

    @property
    def internal_ping_enabled(self) -> str:
        return self._data.get("InternalPingEnabled")

    @property
    def dns_resolution_enabled(self) -> str:
        return self._data.get("DnsResolutionEnabled")

    @property
    def recovery_actions(self) -> int:
        return self._data.get("RecoveryActions")

    @property
    def external_ping_count(self) -> int:
        return self._data.get("ExternalPing_Count")

    @property
    def internal_ping_count(self) -> int:
        return self._data.get("InternalPing_Count")

    @property
    def internal_ping_failed_count(self) -> int:
        return self._data.get("InternalPingFailed_Count")

    @property
    def dns_resolution_test_count(self) -> int:
        return self._data.get("DnsResolutionTest_Count")

    @property
    def reconnection_fix_count(self) -> int:
        return self._data.get("ReconnectionFix_Count")

    @property
    def wifi_recovery_count(self) -> int:
        return self._data.get("HealthCheckWifiRecovery_Count")

    @property
    def wifi_total_uptime(self) -> int:
        return self._data.get("WifiTotalUptime")

    @property
    def wifi_last_uptime(self) -> int:
        return self._data.get("WifiLastUptime")

    @property
    def wifi_total_downtime(self) -> int:
        return self._data.get("WifiTotalDowntime")

    @property
    def wifi_last_downtime(self) -> int:
        return self._data.get("WifiLastDowntime")

    @property
    def wifi_channel_changes(self) -> int:
        return self._data.get("WifiChannelChanges")

    @property
    def wifi_state_no_error_count(self) -> int:
        return self._data.get("WifiStateNoError_Count")

    @property
    def wifi_state_ap_not_found_count(self) -> int:
        return self._data.get("WifiStateApNotFound_Count")

    @property
    def station_disconnected_count(self) -> int:
        return self._data.get("StationConnected_Count")

    @property
    def station_disconnected_count(self) -> int:
        return self._data.get("StationDisconnected_Count")

    @property
    def cloud_connection_drop_count(self) -> int:
        return self._data.get("CloudConnectionDrop_Count")

    @property
    def station_transition_to_connected_count(self) -> int:
        return self._data.get("StationTransitionToConnected_Count")

    @property
    def station_transition_to_disconencted_count(self) -> int:
        return self._data.get("StationTransitionToDisconnected_Count")


class _WiserNetwork:
    """Data structure for network information for a Wiser Hub"""

    def __init__(self, data: dict, wiser_rest_controller: _WiserRestController):
        self._data = data
        self._dhcp_status = data.get("DhcpStatus", {})
        self._network_interface = data.get("NetworkInterface", {})
        self._detected_access_points = []
        self._wiser_rest_controller = wiser_rest_controller

        for detected_network in self._data.get("DetectedAccessPoints", []):
            self._detected_access_points.append(_WiserDetectedNetwork(detected_network))

    @property
    def detected_access_points(self) -> list:
        return self._detected_access_points

    @property
    def dhcp_mode(self) -> str:
        """Get the current dhcp mode of the hub"""
        return self._data.get("NetworkInterface", {}).get("DhcpMode", TEXT_UNKNOWN)

    @property
    def healthckeck_stats(self) -> _WiserNetworkStatistics:
        """Get the network healthcheck stats of the hub"""
        return _WiserNetworkStatistics(self._data.get("HealthCheckStats", {}))

    @property
    def hostname(self) -> str:
        """Get the host name of the hub"""
        return self._data.get("NetworkInterface", {}).get("HostName", TEXT_UNKNOWN)

    @property
    def ip_address(self) -> str:
        """Get the ip address of the hub"""
        if self.dhcp_mode == "Client":
            return self._dhcp_status.get("IPv4Address", TEXT_UNKNOWN)
        else:
            return self._network_interface.get("IPv4HostAddress", TEXT_UNKNOWN)

    @property
    def ip_subnet_mask(self) -> str:
        """Get the subnet mask of the hub"""
        if self.dhcp_mode == "Client":
            return self._dhcp_status.get("IPv4SubnetMask", TEXT_UNKNOWN)
        else:
            return self._network_interface.get("IPv4SubnetMask", TEXT_UNKNOWN)

    @property
    def ip_gateway(self) -> str:
        """Get the default gateway of the hub"""
        if self.dhcp_mode == "Client":
            return self._dhcp_status.get("IPv4DefaultGateway", TEXT_UNKNOWN)
        else:
            return self._network_interface.get("IPv4DefaultGateway", TEXT_UNKNOWN)

    @property
    def ip_primary_dns(self) -> str:
        """Get the primary dns server of the hub"""
        if self.dhcp_mode == "Client":
            return self._dhcp_status.get("IPv4PrimaryDNS", TEXT_UNKNOWN)
        else:
            return self._network_interface.get("IPv4PrimaryDNS", TEXT_UNKNOWN)

    @property
    def ip_secondary_dns(self) -> str:
        """Get the secondary dns server of the hub"""
        if self.dhcp_mode == "Client":
            return self._dhcp_status.get("IPv4SecondaryDNS", TEXT_UNKNOWN)
        else:
            return self._network_interface.get("IPv4SecondaryDNS", TEXT_UNKNOWN)

    @property
    def mac_address(self) -> str:
        """Get the mac address of the hub wifi interface"""
        return self._data.get("MacAddress", TEXT_UNKNOWN)

    @property
    def signal_percent(self) -> int:
        """Get the wifi signal strength percentage"""
        return min(100, int(2 * (self._data.get("RSSI", {}).get("Current", 0) + 100)))

    @property
    def signal_rssi(self) -> int:
        """Get the wifi signal rssi value"""
        return self._data.get("RSSI", {}).get("Current", 0)

    @property
    def signal_rssi_min(self) -> int:
        """Get the wifi signal min rssi value"""
        return self._data.get("RSSI", {}).get("Min", 0)

    @property
    def signal_rssi_max(self) -> int:
        """Get the wifi signal max rssi value"""
        return self._data.get("RSSI", {}).get("Max", 0)

    @property
    def security_mode(self) -> str:
        """Get the wifi security mode"""
        return self._data.get("SecurityMode", TEXT_UNKNOWN)

    @property
    def ssid(self) -> str:
        """Get the ssid of the wifi network the hub is connected to"""
        return self._data.get("SSID", TEXT_UNKNOWN)

    async def connect_to_network(
        self, ssid: str, password: str, channel: int = None, security_mode: str = None
    ):
        """
        Connect hub to wifi network
        param ssid: wifi network ssid
        param password: wifi password
        param channel: wifi channel (optional)
        param security_mode: wifi security mode (optional)
        return: boolean
        """
        cmd_data = {"Enabled": True}
        if ssid and password:
            cmd_data["SSID"] = ssid
            cmd_data["SecurityKey"] = password

        if channel:
            cmd_data["Channel"] = channel

        if security_mode:
            cmd_data["SecurityMode"] = security_mode

        return await self._wiser_rest_controller._send_command(
            f"{WISERHUBNETWORK}/Station", cmd_data
        )
